import json
import os
import random
import boto3
import botocore
import tempfile
import zipfile
import shutil
import time
import functools
import requests
import sys
from zipfile import ZipFile, ZIP_STORED, ZipInfo
import shutil
from aimodelshare.containerization import create_lambda_using_base_image

def create_prediction_api(model_filepath, unique_model_id, model_type, categorical, labels, apiid, custom_libraries, requirements, repo_name="", image_tag=""):

    from zipfile import ZipFile
    import zipfile
    import tempfile
    # create temporary folder
    temp_dir = tempfile.gettempdir()
    import os
    if os.path.exists(os.path.join(temp_dir, 'archive.zip')):
      os.remove(os.path.join(temp_dir, 'archive.zip'))
    else:
      pass
    if os.path.exists(os.path.join(temp_dir, 'archivetest.zip')):
      os.remove(os.path.join(temp_dir, 'archivetest.zip'))
    else:
      pass 	

    if os.path.exists(os.path.join(temp_dir, 'archive2.zip')):
      os.remove(os.path.join(temp_dir, 'archive2.zip'))
    else:
      pass         

    if os.path.exists(os.path.join(temp_dir, 'archive3.zip')):
      os.remove(os.path.join(temp_dir, 'archive3.zip'))
    else:
      pass 
    if os.path.exists(os.path.join(temp_dir,'main.py')):
      os.remove(os.path.join(temp_dir,'main.py'))
    else:
      pass 
    if os.path.exists(os.path.join(temp_dir,'ytest.pkl')):
      os.remove(os.path.join(temp_dir,'ytest.pkl'))
    else:
      pass 
    model_type = model_type.lower()
    categorical = categorical.upper()
    # Wait for 5 seconds to ensure aws iam user on user account has time to load into aws's system
    #time.sleep(5)

    user_session = boto3.session.Session(aws_access_key_id=os.environ.get("AWS_ACCESS_KEY_ID"),
                                          aws_secret_access_key = os.environ.get("AWS_SECRET_ACCESS_KEY"), 
                                          region_name=os.environ.get("AWS_REGION"))
    if(model_type=="neural style transfer"):
            model_layer ="arn:aws:lambda:us-east-1:517169013426:layer:keras_image:1"
            eval_layer ="arn:aws:lambda:us-east-1:517169013426:layer:eval_layer_test:6"
            auth_layer ="arn:aws:lambda:us-east-1:517169013426:layer:aimsauth_layer:2"
    elif model_type=='image' :
            model_layer ="arn:aws:lambda:us-east-1:517169013426:layer:keras_image:1"
            eval_layer ="arn:aws:lambda:us-east-1:517169013426:layer:eval_layer_test:6"
            auth_layer ="arn:aws:lambda:us-east-1:517169013426:layer:aimsauth_layer:2"
    elif model_type=='text':
            model_layer ="arn:aws:lambda:us-east-1:517169013426:layer:tabular_layer:2"
            keras_layer ='arn:aws:lambda:us-east-1:517169013426:layer:keras_preprocesor:1'
            eval_layer ="arn:aws:lambda:us-east-1:517169013426:layer:eval_layer_test:6"
            auth_layer ="arn:aws:lambda:us-east-1:517169013426:layer:aimsauth_layer:2"
    elif model_type == 'tabular' or model_type =='timeseries':
            model_layer ="arn:aws:lambda:us-east-1:517169013426:layer:tabular_cloudpicklelayer:1"
            eval_layer ="arn:aws:lambda:us-east-1:517169013426:layer:eval_layer_test:6"
            auth_layer ="arn:aws:lambda:us-east-1:517169013426:layer:aimsauth_layer:2"
    elif model_type.lower() == 'audio':
            model_layer = "arn:aws:lambda:us-east-1:517169013426:layer:librosa_nosklearn:9"
            eval_layer ="arn:aws:lambda:us-east-1:517169013426:layer:eval_layer_test:6"
            auth_layer ="arn:aws:lambda:us-east-1:517169013426:layer:aimsauth_layer:2"
    elif model_type.lower() == 'video':
            model_layer = "arn:aws:lambda:us-east-1:517169013426:layer:videolayer:3"
            eval_layer ="arn:aws:lambda:us-east-1:517169013426:layer:eval_layer_test:6"
            auth_layer ="arn:aws:lambda:us-east-1:517169013426:layer:aimsauth_layer:2"
    elif model_type.lower() == 'custom':
            model_layer = "arn:aws:lambda:us-east-1:517169013426:layer:videolayer:3"
            eval_layer ="arn:aws:lambda:us-east-1:517169013426:layer:eval_layer_test:6"
            auth_layer ="arn:aws:lambda:us-east-1:517169013426:layer:aimsauth_layer:2"
    else :
        print("no matching model data type to load correct python package zip file (lambda layer)")

    #cloud_layer = "arn:aws:lambda:us-east-1:517169013426:layer:tabular_cloudpicklelayer:1"
    # dill_layer ="arn:aws:lambda:us-east-1:517169013426:layer:dill:3"

    # Update note:  dyndb data to add.  apiname. (include username too)


    account_number = user_session.client(
        'sts').get_caller_identity().get('Account')

    import tempfile
    from zipfile import ZipFile
    import zipfile
    import os

    # create temporary folder
    temp_dir = tempfile.gettempdir()

    try:
      import importlib.resources as pkg_resources

    except ImportError:
      # Try backported to PY<37 `importlib_resources`.
      import importlib_resources as pkg_resources

    from . import main  # relative-import the *package* containing the templates

    from . import custom_approach

    file_objects_folder_path = os.path.join(temp_dir, 'file_objects')

    if model_type.lower() != "custom":  # file_objects already initialized if custom
        if os.path.exists(file_objects_folder_path):
            shutil.rmtree(file_objects_folder_path)
        os.mkdir(file_objects_folder_path)


    # write main handlers
    if(model_type == "neural style transfer"):
            data = pkg_resources.read_text(main, 'nst.txt')
            from string import Template
            t = Template(data)
            newdata = t.substitute(
                bucket_name=os.environ.get("BUCKET_NAME"), unique_model_id=unique_model_id)
    elif model_type == 'text' and categorical == 'TRUE':
            data = pkg_resources.read_text(main, '1.txt')
            from string import Template
            t = Template(data)
            newdata = t.substitute(
                bucket_name=os.environ.get("BUCKET_NAME"), unique_model_id=unique_model_id, labels=labels)
    elif model_type == 'text' and categorical == 'FALSE':
            data = pkg_resources.read_text(main, '1B.txt')
            from string import Template
            t = Template(data)
            newdata = t.substitute(
                bucket_name=os.environ.get("BUCKET_NAME"), unique_model_id=unique_model_id)
    elif model_type == 'image' and categorical == 'TRUE':
            data = pkg_resources.read_text(main, '2.txt')
            from string import Template
            t = Template(data)            
            newdata = t.substitute(
                bucket_name=os.environ.get("BUCKET_NAME"), unique_model_id=unique_model_id, labels=labels)
    elif model_type == 'image' and categorical == 'FALSE':
            data = pkg_resources.read_text(main, '3.txt')
            from string import Template
            t = Template(data)
            newdata = t.substitute(
                bucket_name=os.environ.get("BUCKET_NAME"), unique_model_id=unique_model_id)
    elif all([model_type == 'tabular', categorical == 'TRUE']):
            data = pkg_resources.read_text(main, '4.txt')
            from string import Template
            t = Template(data)
            newdata = t.substitute(
                bucket_name=os.environ.get("BUCKET_NAME"), unique_model_id=unique_model_id, labels=labels)
    elif all([model_type == 'tabular', categorical == 'FALSE']):
            data = pkg_resources.read_text(main, '5.txt')
            from string import Template
            t = Template(data)
            newdata = t.substitute(
                bucket_name=os.environ.get("BUCKET_NAME"), unique_model_id=unique_model_id)
    elif model_type.lower() == 'timeseries' and categorical == 'FALSE':
            data = pkg_resources.read_text(main, '6.txt')
            from string import Template
            t = Template(data)
            newdata = t.substitute(
                bucket_name=os.environ.get("BUCKET_NAME"), unique_model_id=unique_model_id)
    elif model_type.lower() == 'audio' and categorical == 'TRUE':
            data = pkg_resources.read_text(main, '7.txt')
            from string import Template
            t = Template(data)
            newdata = t.substitute(
                bucket_name=os.environ.get("BUCKET_NAME"), unique_model_id=unique_model_id, labels=labels)
    elif model_type.lower() == 'video' and categorical == 'TRUE':
            data = pkg_resources.read_text(main, '8.txt')
            from string import Template
            t = Template(data)
            newdata = t.substitute(
                bucket_name=os.environ.get("BUCKET_NAME"), unique_model_id=unique_model_id, labels=labels)
    elif model_type.lower() == 'custom':
         with open("custom_lambda.py", 'r') as in_file:     
             newdata = in_file.read()

    with open(os.path.join(file_objects_folder_path, 'model.py'), 'w') as file:
        file.write(newdata)

    if(model_type.lower() == 'custom'):
         data = pkg_resources.read_text(custom_approach, 'lambda_function.py')
         with open(os.path.join(file_objects_folder_path, 'lambda_function.py'), 'w') as file:
             file.write(data)
    else:
        data = pkg_resources.read_text(main, 'lambda_function.txt')
        with open(os.path.join(file_objects_folder_path, 'lambda_function.py'), 'w') as file:
            file.write(data)
        
    #with zipfile.ZipFile(os.path.join(temp_dir, 'archive.zip'), 'a') as z:
    #    z.write(os.path.join(temp_dir, 'main.py'), 'main.py')

    # preprocessor upload

    # Upload lambda function zipfile to user's model file folder on s3
    # try:
    #     # This should go to developer's account from my account
    #     s3_client = user_session.client('s3')
    #     s3_client.upload_file(os.path.join(
    #          temp_dir, 'archive.zip'), os.environ.get("BUCKET_NAME"),  unique_model_id+"/"+'archivetest.zip')

    # except Exception as e:
    #     print(e)

    if os.path.exists(os.path.join(temp_dir,'main.py')):
      os.remove(os.path.join(temp_dir,'main.py'))
    else:
      pass  

    if os.path.exists(os.path.join(temp_dir,'archive.zip')):
      os.remove(os.path.join(temp_dir,'archive.zip'))
    else:
      pass   
    
    if any([categorical=="TRUE",categorical=="True",categorical=="true"]):
         task_type="classification"
    elif any([categorical=="FALSE", categorical=="False", categorical=="false"]):
         task_type="regression"
    else:
         task_type="custom"
    # Upload model eval lambda function zipfile to user's model file folder on s3
    # use task_type
    data = pkg_resources.read_text(main, 'eval_lambda.txt')
    from string import Template
    t = Template(data)
    newdata = t.substitute(
        bucket_name=os.environ.get("BUCKET_NAME"), unique_model_id=unique_model_id, task_type=task_type)
    with open(os.path.join(temp_dir, 'main.py'), 'w') as file:
        file.write(newdata)

    with zipfile.ZipFile(os.path.join(temp_dir, 'archive2.zip'), 'a') as z:
        z.write(os.path.join(temp_dir, 'main.py'), 'main.py')

    # preprocessor upload

    # Upload lambda function zipfile to user's model file folder on s3
    try:
        # This should go to developer's account from my account
        s3_client = user_session.client('s3')
        s3_client.upload_file(os.path.join(
            temp_dir, 'archive2.zip'), os.environ.get("BUCKET_NAME"),  unique_model_id+"/"+'archiveeval.zip')

    except Exception as e:
        print(e)

    import os

    if os.path.exists(os.path.join(temp_dir,'main.py')):
      os.remove(os.path.join(temp_dir,'main.py'))
    else:
      pass         

    if os.path.exists(os.path.join(temp_dir,'archive2.zip')):
      os.remove(os.path.join(temp_dir,'archive2.zip'))
    else:
      pass  

    # Upload model eval lambda function zipfile to user's model file folder on s3
    if categorical == 'TRUE':
            newdata = pkg_resources.read_text(main, 'authorization.txt')
            with open(os.path.join(temp_dir, 'main.py'), 'w') as file:
                file.write(newdata)
    elif categorical == 'FALSE':
            newdata = pkg_resources.read_text(main, 'authorization.txt')
            with open(os.path.join(temp_dir, 'main.py'), 'w') as file:
                file.write(newdata)
    with zipfile.ZipFile(os.path.join(temp_dir, 'archive3.zip'), 'a') as z:
        z.write(os.path.join(temp_dir, 'main.py'), 'main.py')

    # preprocessor upload

    # Upload lambda function zipfile to user's model file folder on s3
    try:
        # This should go to developer's account from my account
        s3_client = user_session.client('s3')
        s3_client.upload_file(os.path.join(
            temp_dir, 'archive3.zip'), os.environ.get("BUCKET_NAME"),  unique_model_id+"/"+'archiveauth.zip')

    except Exception as e:
        print(e)

    if model_type.lower() == 'custom':
        s3_client.upload_file(os.path.join(
            temp_dir, 'exampledata.json'), os.environ.get("BUCKET_NAME"),  unique_model_id+"/"+"exampledata.json")
 
    import os
    if os.path.exists(os.path.join(temp_dir, 'archive.zip')):
      os.remove(os.path.join(temp_dir, 'archive.zip'))
    else:
      pass
    if os.path.exists(os.path.join(temp_dir, 'archivetest.zip')):
      os.remove(os.path.join(temp_dir, 'archivetest.zip'))
    else:
      pass 	

    if os.path.exists(os.path.join(temp_dir, 'archive2.zip')):
      os.remove(os.path.join(temp_dir, 'archive2.zip'))
    else:
      pass         

    if os.path.exists(os.path.join(temp_dir, 'archive3.zip')):
      os.remove(os.path.join(temp_dir, 'archive3.zip'))
    else:
      pass 
    if os.path.exists(os.path.join(temp_dir,'main.py')):
      os.remove(os.path.join(temp_dir,'main.py'))
    else:
      pass 


    # Create and/or update roles for lambda function you will create below
    lambdarole1 = {u'Version': u'2012-10-17', u'Statement': [
        {u'Action': u'sts:AssumeRole', u'Effect': u'Allow', u'Principal': {u'Service': u'lambda.amazonaws.com'}}]}

    roles = user_session.client('iam').list_roles()
    
    lambdarolename = 'myService-dev-us-east-1-lambdaRole'+str(random.randint(1, 1000000))
    lambdafxnname = 'modfunction'+str(random.randint(1, 1000000))
    lambdaauthfxnname = 'redisAccess'+str(random.randint(1, 1000000))
    lambdaevalfxnname = 'evalfunction'+str(random.randint(1, 1000000))

    response6 = user_session.resource('iam').create_role(
            AssumeRolePolicyDocument=json.dumps(lambdarole1),
            Path='/',
            RoleName=lambdarolename,
        )
    response6_2 = user_session.client('iam').put_role_policy(
            PolicyDocument='{"Version":"2012-10-17","Statement":[{"Action": ["logs:CreateLogGroup"],"Resource": ["arn:aws:logs:us-east-1:'+account_number+':*"],"Effect": "Allow"},{"Action": ["logs:CreateLogStream"], "Resource": ["arn:aws:logs:us-east-1:'+account_number+':log-group:/aws/lambda/'+lambdafxnname +
            ':*"],"Effect": "Allow"},{"Action": ["logs:PutLogEvents"],"Resource": ["arn:aws:logs:us-east-1:'+account_number+':log-group:/aws/lambda/' +
            lambdafxnname +
            ':*:*"],"Effect": "Allow"},{"Action": ["s3:GetObject"],"Resource": ["arn:aws:s3:::' +
            os.environ.get("BUCKET_NAME")+'/*"],"Effect": "Allow"}]}',
            PolicyName='S3AccessandcloudwatchlogPolicy'+str(random.randint(1, 1000000)),
            RoleName=lambdarolename,
        )

    response6_2 = user_session.client('iam').put_role_policy(
            PolicyDocument='{"Version":"2012-10-17","Statement":[{"Action": ["logs:CreateLogStream"], "Resource": ["arn:aws:logs:us-east-1:'+account_number+':log-group:/aws/lambda/'+lambdaauthfxnname +
            ':*"],"Effect": "Allow"},{"Action": ["logs:PutLogEvents"],"Resource": ["arn:aws:logs:us-east-1:'+account_number+':log-group:/aws/lambda/' +
            lambdaauthfxnname +
            ':*:*"],"Effect": "Allow"},{"Action": ["s3:GetObject"],"Resource": ["arn:aws:s3:::' +
            os.environ.get("BUCKET_NAME")+'/*"],"Effect": "Allow"}]}',
            PolicyName='S3AccessandcloudwatchlogPolicy'+str(random.randint(1, 1000000)),
            RoleName=lambdarolename,
        )

    response6_2 = user_session.client('iam').put_role_policy(
            PolicyDocument='{"Version":"2012-10-17","Statement":[{"Action": ["logs:CreateLogStream"], "Resource": ["arn:aws:logs:us-east-1:'+account_number+':log-group:/aws/lambda/'+lambdaevalfxnname +
            ':*"],"Effect": "Allow"},{"Action": ["logs:PutLogEvents"],"Resource": ["arn:aws:logs:us-east-1:'+account_number+':log-group:/aws/lambda/' +
            lambdaevalfxnname +
            ':*:*"],"Effect": "Allow"},{"Action": ["s3:ListBucket"],"Resource": ["arn:aws:s3:::' +
            os.environ.get("BUCKET_NAME")+'"],"Effect": "Allow"},{"Action": ["s3:GetObject"],"Resource": ["arn:aws:s3:::' +
            os.environ.get("BUCKET_NAME")+'/*"],"Effect": "Allow"},{"Action": ["s3:PutObject"],"Resource": ["arn:aws:s3:::' +
            os.environ.get("BUCKET_NAME")+'/*"],"Effect": "Allow"}]}',
            PolicyName='S3AccessandcloudwatchlogPolicy'+str(random.randint(1, 1000000)),
            RoleName=lambdarolename,
        )

    lambdaclient = user_session.client('lambda')

    ##!!!  this is f'd.  looks like aishwarya hasn't fixed the reference to other layers!  It's a good start, but we should add a subfunction to return
    ## the correct layer from an externally stored list (save arns to a github repo and allow them to be imported here at some point.)
    layers = []
    layers.append(model_layer)
    if model_type =='text':
        layers.append(keras_layer)

    # if model_type=='sklearn_text' or  model_type=='keras_text' or model_type=='flubber_text' or model_type =='text':
    # layers.append(keras_layer)

    #response6 = lambdaclient.create_function(FunctionName=lambdafxnname, Runtime='python3.6', Role='arn:aws:iam::'+account_number+':role/'+lambdarolename, Handler='main.handler',
    #                                          Code={
    #                                              'S3Bucket': os.environ.get("BUCKET_NAME"),
    #                                              'S3Key':  unique_model_id+"/"+'archivetest.zip'
    #                                          }, Timeout=10, MemorySize=512, Layers=layers)  # ADD ANOTHER LAYER ARN .. THE ONE SPECIFIC TO MODEL TYPE


    ### Progress Update #3/6 {{{
    sys.stdout.write('\r')
    sys.stdout.write("[============                         ] Progress: 40% - Creating custom containers...                        ")
    sys.stdout.flush()
    # }}}
    
    if(any([custom_libraries=='FALSE',custom_libraries=='false'])):
        response6 = create_lambda_using_base_image(user_session, os.getenv("BUCKET_NAME"), file_objects_folder_path, lambdafxnname, apiid, repo_name, image_tag, 3072, 90)
    elif(any([custom_libraries=='TRUE',custom_libraries=='true'])):

        requirements = requirements.split(",")
        for i in range(len(requirements)):
            requirements[i] = requirements[i].strip(" ")

        with open(os.path.join(file_objects_folder_path, 'requirements.txt'), 'a') as f:
            for lib in requirements:
                f.write('%s\n' % lib)
        
        requirements_file_path = os.path.join(file_objects_folder_path, 'requirements.txt')

        from aimodelshare import deploy_container
        #response6 = deploy_lambda_using_sam(user_session, os.getenv("BUCKET_NAME"), requirements, file_objects_folder_path, lambdafxnname, apiid, 1024, 90, "3.7")
        response6 = deploy_container(account_number, os.environ.get("AWS_REGION"), user_session, lambdafxnname, file_objects_folder_path,requirements_file_path,apiid)

    response6evalfxn = lambdaclient.create_function(FunctionName=lambdaevalfxnname, Runtime='python3.7', Role='arn:aws:iam::'+account_number+':role/'+lambdarolename, Handler='main.handler',
                                          Code={
                                              'S3Bucket': os.environ.get("BUCKET_NAME"),
                                              'S3Key':  unique_model_id+"/"+'archiveeval.zip'
                                          }, Timeout=90, MemorySize=2048, Layers=[eval_layer,auth_layer])  # ADD ANOTHER LAYER ARN .. THE ONE SPECIFIC TO MODEL TYPE

    response6authfxn = lambdaclient.create_function(FunctionName=lambdaauthfxnname, Runtime='python3.7', Role='arn:aws:iam::'+account_number+':role/'+lambdarolename, Handler='main.handler',
                                          Code={
                                              'S3Bucket': os.environ.get("BUCKET_NAME"),
                                              'S3Key':  unique_model_id+"/"+'archiveauth.zip'
                                          }, Timeout=10, MemorySize=512, Layers=[auth_layer])  # ADD ANOTHER LAYER ARN .. THE ONE SPECIFIC TO MODEL TYPE

    ### Progress Update #4/6 {{{
    sys.stdout.write('\r')
    sys.stdout.write("[==========================           ] Progress: 75% - Deploying prediction API...                          ")
    sys.stdout.flush()
    # }}}


    #add create api about here
    #TODO: 
    api_name = 'modapi'+str(random.randint(1, 1000000))	

    # Update note:  change apiname in apijson from modapi890799 to randomly generated apiname?  or aimodelshare generic name?

    user_client = boto3.client('apigateway', aws_access_key_id=str(
        os.environ.get("AWS_ACCESS_KEY_ID")), aws_secret_access_key=str(os.environ.get("AWS_SECRET_ACCESS_KEY")), region_name=str(os.environ.get("AWS_REGION")))

    api_id = apiid

    # Update note:  dyndb data to add.  api_id and resourceid "Resource": "arn:aws:execute-api:us-east-1:517169013426:iu3q9io652/prod/OPTIONS/m"

    # Update note:  dyndb data to add.  api_id and resourceid "Resource": "arn:aws:execute-api:us-east-1:517169013426:iu3q9io652/prod/OPTIONS/m"


    response3 = user_client.get_resources(restApiId=api_id)


    resourceidlist=response3['items']

    # Python3 code to iterate over a list 
    api_id_data = {}   
    # Using for loop 

    for i in resourceidlist: 
        api_id_data.update({i['path']: i['id']})

    resource_id_parent=api_id_data['/']

    resource_id=api_id_data['/m']

    resource_id_eval=api_id_data['/eval']


    # NEXT: update permissions

    fxn_list = lambdaclient.list_functions()

    stmt_id = 'apigateway-prod-'+str(random.randint(1, 1000000))
    # upload authfxn code first

    response7_1 = lambdaclient.add_permission(
            FunctionName=lambdaauthfxnname,
            StatementId=stmt_id,
            Action='lambda:InvokeFunction',
            Principal='apigateway.amazonaws.com',
            SourceArn='arn:aws:execute-api:us-east-1:'+account_number+":"+api_id+'/*/*',
    )
    # Update note:  dyndb data to add.  lambdafxnname

    # change api name below?

    time.sleep(15)

    response7 = lambdaclient.add_permission(
        FunctionName=lambdafxnname,
        StatementId='apigateway-prod-2',
        Action='lambda:InvokeFunction',
        Principal='apigateway.amazonaws.com',
        SourceArn='arn:aws:execute-api:us-east-1:' +
        account_number+":"+api_id+'/*/POST/m',
    )

    response8 = lambdaclient.add_permission(
        FunctionName=lambdafxnname,
        StatementId='apigateway-test-2',
        Action='lambda:InvokeFunction',
        Principal='apigateway.amazonaws.com',
        SourceArn='arn:aws:execute-api:us-east-1:' +
        account_number+":"+api_id+'/*/POST/m',
    )

    response7 = lambdaclient.add_permission(
        FunctionName=lambdaevalfxnname,
        StatementId='apigateway-prod-3',
        Action='lambda:InvokeFunction',
        Principal='apigateway.amazonaws.com',
        SourceArn='arn:aws:execute-api:us-east-1:' +
        account_number+":"+api_id+'/*/POST/eval',
    )

    response8 = lambdaclient.add_permission(
        FunctionName=lambdaevalfxnname,
        StatementId='apigateway-test-3',
        Action='lambda:InvokeFunction',
        Principal='apigateway.amazonaws.com',
        SourceArn='arn:aws:execute-api:us-east-1:' +
        account_number+":"+api_id+'/*/POST/eval',
    )

    # Create and or update lambda and apigateway gateway roles


    lambdarole2 = {u'Version': u'2012-10-17', u'Statement': [{u'Action': u'sts:AssumeRole', u'Principal': {
        u'Service': [u'lambda.amazonaws.com', u'apigateway.amazonaws.com']}, u'Effect': u'Allow', u'Sid': u''}]}

    if str(roles['Roles']).find("lambda_invoke_function_assume_apigw_role") > 0:
        None
    else:
        response10 = user_session.resource('iam').create_role(
            AssumeRolePolicyDocument=json.dumps(lambdarole2),
            Path='/',
            RoleName='lambda_invoke_function_assume_apigw_role',
        )
        time.sleep(10)

    response11 = user_session.client('apigateway').put_integration(
        restApiId=api_id,
        resourceId=resource_id,
        httpMethod='POST',
        type='AWS_PROXY',
        integrationHttpMethod='POST',
        uri='arn:aws:apigateway:us-east-1:lambda:path/2015-03-31/functions/arn:aws:lambda:us-east-1:' +
        account_number+':function:'+lambdafxnname+'/invocations',
        credentials='arn:aws:iam::'+account_number+':role/lambda_invoke_function_assume_apigw_role')
    response11_1 = user_session.client('apigateway').put_integration(
        restApiId=api_id,
        resourceId=resource_id,
        httpMethod='OPTIONS',
        type='MOCK',
        requestTemplates={
            'application/json': '{"statusCode": 200}'
        },
        integrationHttpMethod='OPTIONS',
        uri='arn:aws:apigateway:us-east-1:lambda:path/2015-03-31/functions/arn:aws:lambda:us-east-1:' +
        account_number+':function:'+lambdafxnname+'/invocations',
        credentials='arn:aws:iam::'+account_number+':role/lambda_invoke_function_assume_apigw_role')
    response11_1B = user_session.client('apigateway').put_integration(
        restApiId=api_id,
        resourceId=resource_id,
        httpMethod='OPTIONS',
        type='MOCK',
        requestTemplates={
            'application/json': '{"statusCode": 200}'
        }
    )
    response11_1C = user_session.client('apigateway').put_integration_response(
        restApiId=api_id,
        resourceId=resource_id,
        httpMethod='OPTIONS',
        statusCode='200',
        responseParameters={
            'method.response.header.Access-Control-Allow-Headers': '\'Content-Type,X-Amz-Date,authorizationToken,Access-Control-Allow-Origin,X-Api-Key,X-Amz-Security-Token\'',
            'method.response.header.Access-Control-Allow-Methods': '\'POST,OPTIONS\'',
            'method.response.header.Access-Control-Allow-Origin': '\'*\''
        },
        responseTemplates={
            'application/json': '{"statusCode": 200}'
        }
    )
    response11_2 = user_session.client('iam').put_role_policy(
        PolicyDocument='{"Version":"2012-10-17","Statement":{"Effect":"Allow","Action":"lambda:InvokeFunction","Resource":"*"}}',
        PolicyName='invokelambda',
        RoleName='lambda_invoke_function_assume_apigw_role',
    )
    response12_1 = user_client.update_rest_api(
        restApiId=api_id,
        patchOperations=[{
            "op": "replace",
            "path": "/policy",
            "value": '{"Version": "2012-10-17","Statement": [{"Effect": "Allow","Principal": "*","Action": "execute-api:Invoke","Resource": "arn:aws:execute-api:'+os.environ.get("AWS_REGION")+':'+account_number+':'+api_id+'/prod/OPTIONS/*"}]}'
        }, ]
    )

    # start here to update eval fxn integration with api resource_id_eval, lambdaevalfxnname
    response11 = user_session.client('apigateway').put_integration(
        restApiId=api_id,
        resourceId=resource_id_eval,
        httpMethod='POST',
        type='AWS_PROXY',
        integrationHttpMethod='POST',
        uri='arn:aws:apigateway:us-east-1:lambda:path/2015-03-31/functions/arn:aws:lambda:us-east-1:' +
        account_number+':function:'+lambdaevalfxnname+'/invocations',
        credentials='arn:aws:iam::'+account_number+':role/lambda_invoke_function_assume_apigw_role')
    response11_1 = user_session.client('apigateway').put_integration(
        restApiId=api_id,
        resourceId=resource_id_eval,
        httpMethod='OPTIONS',
        type='MOCK',
        requestTemplates={
            'application/json': '{"statusCode": 200}'
        },
        integrationHttpMethod='OPTIONS',
        uri='arn:aws:apigateway:us-east-1:lambda:path/2015-03-31/functions/arn:aws:lambda:us-east-1:' +
        account_number+':function:'+lambdaevalfxnname+'/invocations',
        credentials='arn:aws:iam::'+account_number+':role/lambda_invoke_function_assume_apigw_role')
    response11_1B = user_session.client('apigateway').put_integration(
        restApiId=api_id,
        resourceId=resource_id_eval,
        httpMethod='OPTIONS',
        type='MOCK',
        requestTemplates={
            'application/json': '{"statusCode": 200}'
        }
    )
    response11_1C = user_session.client('apigateway').put_integration_response(
        restApiId=api_id,
        resourceId=resource_id_eval,
        httpMethod='OPTIONS',
        statusCode='200',
        responseParameters={
            'method.response.header.Access-Control-Allow-Headers': '\'Content-Type,X-Amz-Date,authorizationToken,Access-Control-Allow-Origin,X-Api-Key,X-Amz-Security-Token\'',
            'method.response.header.Access-Control-Allow-Methods': '\'POST,OPTIONS\'',
            'method.response.header.Access-Control-Allow-Origin': '\'*\''
        },
        responseTemplates={
            'application/json': '{"statusCode": 200}'
        }
    )
    response11_2 = user_session.client('iam').put_role_policy(
        PolicyDocument='{"Version":"2012-10-17","Statement":{"Effect":"Allow","Action":"lambda:InvokeFunction","Resource":"*"}}',
        PolicyName='invokelambda',
        RoleName='lambda_invoke_function_assume_apigw_role',
    )
        
    response12_1 = user_client.update_rest_api(
        restApiId=api_id,
        patchOperations=[{
            "op": "replace",
            "path": "/policy",
            "value": '{"Version": "2012-10-17","Statement": [{"Effect": "Allow","Principal": "*","Action": "execute-api:Invoke","Resource": "arn:aws:execute-api:'+ os.environ.get("AWS_REGION") +':'+account_number+':'+api_id+'/prod/OPTIONS/*"}]}'
        }, ]
    )

    responseauthfxnapigateway = user_session.client('apigateway').create_authorizer(
        restApiId=api_id,
        name='aimscustomauthfxn',
        type='TOKEN',
        authorizerUri="arn:aws:apigateway:"+ os.environ.get("AWS_REGION") +":lambda:path/2015-03-31/functions/arn:aws:lambda:"+os.environ.get("AWS_REGION")+":"+account_number+":function:"+lambdaauthfxnname+"/invocations",
        identitySource="method.request.header.authorizationToken",
        authorizerResultTtlInSeconds=0
    )

    responseauthfxnapigateway = user_session.client('apigateway').get_authorizers(
        restApiId=api_id
    )

    authorizerid=responseauthfxnapigateway['items'][0]['id']

    stmt_idauth = 'apigateway-prod-'+str(random.randint(1, 1000000))
    response70 = user_session.client('lambda').add_permission(
        FunctionName=lambdaauthfxnname,
        StatementId=stmt_idauth,
        Action='lambda:InvokeFunction',
        Principal='apigateway.amazonaws.com',
        SourceArn='arn:aws:execute-api:'+os.environ.get("AWS_REGION")+':' +
        account_number+':'+api_id+'/authorizers/'+authorizerid,
    )
    
    response_modmthd_addauth = user_session.client('apigateway').update_method(
        restApiId=api_id,
        resourceId=resource_id,
        httpMethod='POST',
        patchOperations=[
            {
                'op': 'replace',
                'path': '/authorizationType',
                'value': 'CUSTOM',
                'from': 'NONE'
            },
            {
                'op': 'replace',
                'path': '/authorizerId',
                'value': authorizerid
            }])

    response_evalmthd_addauth = user_session.client('apigateway').update_method(
        restApiId=api_id,
        resourceId=resource_id_eval,
        httpMethod='POST',
        patchOperations=[
            {
                'op': 'replace',
                'path': '/authorizationType',
                'value': 'CUSTOM',
                'from': 'NONE'
            },
            {
                'op': 'replace',
                'path': '/authorizerId',
                'value': authorizerid
            }])
    response12 = user_session.client('apigateway').create_deployment(
        restApiId=api_id,
        stageName='prod')


    result = 'https://'+api_id + '.execute-api.'+os.environ.get("AWS_REGION")+'.amazonaws.com/prod/m'

    return {"statusCode": 200,
            "headers": {
                "Access-Control-Allow-Origin": "*",
                "Access-Control-Allow-Credentials": True
            },
            "body": json.dumps(result)}



def get_api_json():
    apijson = '''
        {
          "openapi": "3.0.1",
          "info": {
            "title": "modapi36146",
            "description": "This is a copy of my first API",
            "version": "2020-05-12T21:25:38Z"
          },
          "servers": [
            {
              "url": "https://8nee9nskdb.execute-api.us-east-1.amazonaws.com/{basePath}",
              "variables": {
                "basePath": {
                  "default": "/prod"
                }
              }
            }
          ],
          "paths": {
            "/eval": {
              "post": {
                "responses": {
                  "200": {
                    "description": "200 response",
                    "headers": {
                      "Access-Control-Allow-Origin": {
                        "schema": {
                          "type": "string"
                        }
                      }
                    },
                    "content": {
                      "application/json": {
                        "schema": {
                          "$ref": "#/components/schemas/outputmodel"
                        }
                      }
                    }
                  }
                }
              },
              "options": {
                "responses": {
                  "200": {
                    "description": "200 response",
                    "headers": {
                      "Access-Control-Allow-Origin": {
                        "schema": {
                          "type": "string"
                        }
                      },
                      "Access-Control-Allow-Methods": {
                        "schema": {
                          "type": "string"
                        }
                      },
                      "Access-Control-Allow-Headers": {
                        "schema": {
                          "type": "string"
                        }
                      }
                    },
                    "content": {
                      "application/json": {
                        "schema": {
                          "$ref": "#/components/schemas/Empty"
                        }
                      }
                    }
                  }
                }
              }
            },
            "/m": {
              "post": {
                "responses": {
                  "200": {
                    "description": "200 response",
                    "headers": {
                      "Access-Control-Allow-Origin": {
                        "schema": {
                          "type": "string"
                        }
                      }
                    },
                    "content": {
                      "application/json": {
                        "schema": {
                          "$ref": "#/components/schemas/outputmodel"
                        }
                      }
                    }
                  }
                }
              },
              "options": {
                "responses": {
                  "200": {
                    "description": "200 response",
                    "headers": {
                      "Access-Control-Allow-Origin": {
                        "schema": {
                          "type": "string"
                        }
                      },
                      "Access-Control-Allow-Methods": {
                        "schema": {
                          "type": "string"
                        }
                      },
                      "Access-Control-Allow-Headers": {
                        "schema": {
                          "type": "string"
                        }
                      }
                    },
                    "content": {
                      "application/json": {
                        "schema": {
                          "$ref": "#/components/schemas/Empty"
                        }
                      }
                    }
                  }
                }
              }
            }
          },
          "components": {
            "schemas": {
              "Empty": {
                "title": "Empty Schema",
                "type": "object"
              },
              "outputmodel": {
                "title": "Output",
                "type": "object",
                "properties": {
                  "body": {
                    "type": "string"
                  }
                }
              }
            }
          },
          "x-amazon-apigateway-policy": {
            "Version": "2012-10-17",
            "Statement": [
              {
                "Effect": "Allow",
                "Principal": "*",
                "Action": "execute-api:Invoke",
                "Resource": "arn:aws:execute-api:us-east-1:517169013426:8nee9nskdb/prod/OPTIONS/*"
              }
            ]
          }
        }
                '''
    return apijson



    
def delete_deployment(apiurl):
    """
    apiurl: string of API URL the user wishes to delete

    WARNING: User must supply high-level credentials in order to delete an API. 
    """
    from aimodelshare.aws import run_function_on_lambda
    
    # Provide Warning & Have user confirm deletion 
    print("Running this function will permanently delete all resources tied to this deployment, \n including the eval lambda and all models submitted to the model competition.\n")
    confirmation = input(prompt="To confirm, type 'permanently delete':")
    if confirmation.lower() == "permanently delete" or confirmation.lower() == "'permanently delete'":
        pass
    else:
        return print("'Delete Deployment' unsuccessful: operation cancelled by user.")

    # Confirm that creds are loaded, print warning if not
    if all(["AWS_ACCESS_KEY_ID" in os.environ, 
            "AWS_SECRET_ACCESS_KEY" in os.environ,
            "AWS_REGION" in os.environ,
           "username" in os.environ, 
           "password" in os.environ]):
        pass
    else:
        return print("'Delete Deployment' unsuccessful. Please provide credentials with set_credentials().")
    
    # get api_id from apiurl
    api_url_trim = apiurl.split('https://')[1]
    api_id = api_url_trim.split(".")[0]

    # Create User Session
    user_sess = boto3.session.Session(aws_access_key_id=os.environ.get('AWS_ACCESS_KEY_ID'), 
                                      aws_secret_access_key=os.environ.get('AWS_SECRET_ACCESS_KEY'), 
                                      region_name=os.environ.get('AWS_REGION'))
    
    s3 = user_sess.resource('s3')

    # Get bucket and model_id subfolder for user based on apiurl {{{
    response, error = run_function_on_lambda(
        apiurl, **{"delete": "FALSE", "versionupdateget": "TRUE"}
    )
    if error is not None:
        raise error

    _, api_bucket, model_id = json.loads(response.content.decode("utf-8"))
    # }}} 

    #Confirm username in bucket name
    ## TODO: Update this check to more secure process
    if os.environ.get("username").lower() in api_bucket:
        pass
    else:
        print("Permission denied. Please provide credentials that allow administrator access to this api.")
        return

    # delete s3 folder
    bucket = s3.Bucket(api_bucket)
    bucket.objects.filter(Prefix= model_id+'/').delete() 

    # get api resources
    api = user_sess.client('apigateway')
    resources = api.get_resources(
        restApiId=api_id
        )

    #get lambda arns 
    lambda_arns = list()
    for i in range(len(resources['items'])):
        if len(resources['items'][i]) > 2: 
            resource_id = resources['items'][i]['id']
            integration = api.get_integration(
                restApiId=api_id,
                resourceId=resource_id,
                httpMethod='POST'
            )
            uri=integration['uri']
            ans1=uri.split('functions/')
            lambda_arn = ans1[1].split('/invocations')[0]
            lambda_arns.append(lambda_arn)
    else: 
        pass 

    # get authorizer arn 
    authorizers = api.get_authorizers(
                  restApiId=api_id
                  )

    authorizer_full_arn=authorizers['items'][0]['authorizerUri']
    ans1=authorizer_full_arn.split('functions/')
    auth_arn = ans1[1].split('/invocations')[0]
    lambda_arns.append(auth_arn)

    # delete lambdas & authorizer
    client = boto3.client('lambda', region_name=os.environ.get("AWS_REGION"))
    for arn in lambda_arns:
        lambda_response = client.delete_function(
            FunctionName = arn
        )

    # delete api
    client = boto3.client('apigateway', region_name=os.environ.get("AWS_REGION"))
    api_response = client.delete_rest_api(
        restApiId=api_id
    )

    # delete api page on front end
    bodydata = {'apiurl': apiurl,
                'delete': "TRUE"
                }
    headers_with_authentication = {'Content-Type': 'application/json', 'authorizationToken': os.environ.get("JWT_AUTHORIZATION_TOKEN"), 'Access-Control-Allow-Headers':
                                   'Content-Type,X-Amz-Date,authorizationToken,Access-Control-Allow-Origin,X-Api-Key,X-Amz-Security-Token,Authorization', 'Access-Control-Allow-Origin': '*'}

    requests.post("https://bhrdesksak.execute-api.us-east-1.amazonaws.com/dev/modeldata",
                  json=bodydata, headers=headers_with_authentication)
    
    # delete competition posting
    bodydata = {"apiurl": apiurl,
                "delete":"TRUE"
                                }

    # Get the response
    headers_with_authentication = {'Content-Type': 'application/json', 'authorizationToken': os.environ.get("AWS_TOKEN"), 'Access-Control-Allow-Headers':
                                    'Content-Type,X-Amz-Date,authorizationToken,Access-Control-Allow-Origin,X-Api-Key,X-Amz-Security-Token,Authorization', 'Access-Control-Allow-Origin': '*'}
    # competitiondata lambda function invoked through below url to update model submissions and contributors
    requests.post("https://o35jwfakca.execute-api.us-east-1.amazonaws.com/dev/modeldata",
                  json=bodydata, headers=headers_with_authentication)

    return "API deleted successfully."


__all__ = [
    get_api_json,
    create_prediction_api,
    delete_deployment
]



# import packages 




class ModelPlayground:
    """
    Parameters:
    model_type :  string 
                  values - [ 'Text' , 'Image' , 'Tabular' , 'Timeseries' ] 
                  type of model data     
    classification:    bool, default=True
                        True [DEFAULT] if model is of Classification type with categorical target variables
                        False if model is of Regression type with continuous target variables
    private :   bool, default = False
                    True if model and its corresponding data is not public
                    False [DEFAULT] if model and its corresponding data is public 
    """
    def __init__(self, model_type=None, classification=None, private=None, playground_url=None):
        # confirm correct args are provided
        if playground_url != None or all([model_type !=None, classification !=None, private!=None]):
            pass
        elif playground_url == None and any([model_type ==None, classification ==None, private==None]):
            return print("Error. To instantiate a ModelPlayground instance, please provide either a playground_url or \n the model_type, classification, and private arguments.")
        
        self.model_type = model_type
        self.categorical = classification 
        self.private = private
        self.playground_url = playground_url
    
    
    def __str__(self):
        return f"ModelPlayground instance of model type: {self.model_type}, classification: {self.categorical},  private: {self.private}"
    
    def deploy(self, model_filepath, preprocessor_filepath, y_train, example_data=None, custom_libraries = "FALSE"):
        """
        Launches a live prediction REST API for deploying ML models using model parameters and user credentials, provided by the user
        Inputs : 7
        Output : model launched to an API
                detaled API info printed out
        -----------
        Parameters 
        
        model_filepath :  string ends with '.onnx'
                          value - Absolute path to model file 
                          [REQUIRED] to be set by the user
                          .onnx is the only accepted model file extension
                          "example_model.onnx" filename for file in directory.
                          "/User/xyz/model/example_model.onnx" absolute path to model file from local directory
        preprocessor_filepath:  string
                                value - absolute path to preprocessor file 
                                [REQUIRED] to be set by the user
                                "./preprocessor.zip" 
                                searches for an exported zip preprocessor file in the current directory
                                file is generated using export_preprocessor function from the AI Modelshare library  
        y_train : training labels for classification models.
                  [REQUIRED] for classification type models
                  expects pandas dataframe of one hot encoded y train data
        custom_libraries:   string
                    "TRUE" if user wants to load custom Python libraries to their prediction runtime
                    "FALSE" if user wishes to use AI Model Share base libraries including latest versions of most common ML libs.
        example_data:  pandas DataFrame (for tabular & text data) OR filepath as string (image, audio, video data)
                      tabular data - pandas DataFrame in same structure expected by preprocessor function
                      other data types - absolute path to folder containing example data
                                          (first five files with relevent file extensions will be accepted)
                      [REQUIRED] for tabular data
        -----------
        Returns
        print_api_info : prints statements with generated live prediction API details
                        also prints steps to update the model submissions by the user/team
        """
        from aimodelshare.generatemodelapi import model_to_api
        self.playground_url = model_to_api(model_filepath=model_filepath, 
                                      model_type = self.model_type, 
                                      private = self.private, 
                                      categorical = self.categorical,
                                      y_train = y_train, 
                                      preprocessor_filepath = preprocessor_filepath, 
                                      example_data = example_data,
                                      custom_libraries = custom_libraries)
        #remove extra quotes
        self.playground_url = self.playground_url[1:-1]
    
    def create_competition(self, data_directory, y_test, email_list = [], public=False):
        """
        Creates a model competition for a deployed prediction REST API
        Inputs : 4
        Output : Create ML model competition and allow authorized users to submit models to resulting leaderboard/competition
        
        ---------
        Parameters
        y_test :  list of y values for test data used to generate metrics from predicted values from X test data submitted via the submit_model() function
                [REQUIRED] to generate eval metrics in competition leaderboard
                                
        data_directory : folder storing training data and test data (excluding Y test data)
        email_list: [OPTIONAL] list of comma separated emails for users who are allowed to submit models to competition.  Emails should be strings in a list.
        public: [REQUIRED] True/false. Defaults to False.  If True, competition is public and ANY AIMODELSHARE USER CAN SUBMIT MODELS.  USE WITH CAUTION b/c one model and 
                           one preprocessor file will be be saved to your AWS S3 folder for each model submission.
        
        
        ---------
        Returns
        finalmessage : Information such as how to submit models to competition
        
        """
        from aimodelshare.generatemodelapi import create_competition as to_competition
        competition = to_competition(self.playground_url, 
                                    data_directory, 
                                    y_test, 
                                    email_list, public)
        return competition
        
    def submit_model(self, model_filepath, preprocessor_filepath, prediction_submission):
        """
        Submits model/preprocessor to machine learning competition using live prediction API url generated by AI Modelshare library
        The submitted model gets evaluated and compared with all existing models and a leaderboard can be generated 
        ---------------
        Parameters:
        model_filepath:  string ends with '.onnx'
                    value - Absolute path to model file [REQUIRED] to be set by the user
                    .onnx is the only accepted model file extension
                    "example_model.onnx" filename for file in directory.
                    "/User/xyz/model/example_model.onnx" absolute path to model file from local directory
        preprocessor_filepath:   string,default=None
                        value - absolute path to preprocessor file 
                        [REQUIRED] to be set by the user
                        "./preprocessor.zip" 
                        searches for an exported zip preprocessor file in the current directory
                        file is generated from preprocessor module using export_preprocessor function from the AI Modelshare library 
        prediction_submission: [REQUIRED] list of predictions from X test data that will be used to evaluate model prediction error against y test data.
                         Use mycompetition.inspect_y_test() to view example of list expected by competition.
        -----------------
        Returns
        response:   Model version if the model is submitted sucessfully
                    error  if there is any error while submitting models
        
        """
        from aimodelshare.model import submit_model as submit
        submission = submit(model = model_filepath, 
                            apiurl = self.playground_url,
                            prediction_submission = prediction_submission, 
                            preprocessor = preprocessor_filepath)
        return submission
        

    
    def update_runtime_model(self, model_version=None): 
        from aimodelshare.model import update_runtime_model as update
        update = update(apiurl = self.playground_url, model_version = model_version)
        return update
        
    def instantiate_model(self, version=None, trained=False): 
        from aimodelshare.aimsonnx import instantiate_model
        model = instantiate_model(apiurl=self.playground_url, trained=trained, version=version)
        return model
    
    def delete_deployment(self, playground_url=None):
        """
        Delete all components of a Model Playground, including: AWS s3 bucket & contents,
        attached competitions, prediction REST API, and interactive Model Playground web dashboard.
        ---------------
        playground_url: string of API URL the user wishes to delete
        WARNING: User must supply high-level credentials in order to delete an API. 
        """
        from aimodelshare.api import delete_deployment
        if playground_url == None:
            playground_url = self.playground_url
        deletion = delete_deployment(apiurl = playground_url)
        return deletion




class Competition: 
    def __init__(self, playground_url):
        self.playground_url = playground_url
    
    def __str__(self):
        return f"Competition class instance for playground: {self.playground_url}"
        
    def submit_model(self, model_filepath, preprocessor_filepath, prediction_submission, sample_data=None):
        """
        Submits model/preprocessor to machine learning competition using live prediction API url generated by AI Modelshare library
        The submitted model gets evaluated and compared with all existing models and a leaderboard can be generated 
        ---------------
        Parameters:
        model_filepath:  string ends with '.onnx'
                    value - Absolute path to model file [REQUIRED] to be set by the user
                    .onnx is the only accepted model file extension
                    "example_model.onnx" filename for file in directory.
                    "/User/xyz/model/example_model.onnx" absolute path to model file from local directory
        prediction_submission:   one hot encoded y_pred
                        value - predictions for test data
                        [REQUIRED] for evaluation metriicts of the submitted model
        preprocessor_filepath:   string,default=None
                        value - absolute path to preprocessor file 
                        [REQUIRED] to be set by the user
                        "./preprocessor.zip" 
                        searches for an exported zip preprocessor file in the current directory
                        file is generated from preprocessor module using export_preprocessor function from the AI Modelshare library 
        -----------------
        Returns
        response:   Model version if the model is submitted sucessfully
                    error  if there is any error while submitting models
        
        """
        from aimodelshare.model import submit_model as submit
        submission = submit(model_filepath = model_filepath, 
                              apiurl = self.playground_url,
                              prediction_submission = prediction_submission, 
                              preprocessor = preprocessor_filepath)
        return submission
        
    def instantiate_model(self, version=None, trained=False): 
        from aimodelshare.aimsonnx import instantiate_model
        model = instantiate_model(apiurl=self.playground_url, trained=trained, version=version)
        return model

    def inspect_model(self, version=None):
        from aimodelshare.aimsonnx import inspect_model
        inspect_pd = inspect_model(apiurl=self.playground_url, version=version)
        return inspect_pd

    def compare_models(self, version_list="None", by_model_type=None, best_model=None, verbose=3):
        from aimodelshare.aimsonnx import compare_models as compare
        data = compare(apiurl = self.playground_url, 
                      version_list = version_list, 
                      by_model_type = by_model_type,
                      best_model = best_model, 
                      verbose = verbose)
        return data

    def stylize_compare(self, compare_dict):
        from aimodelshare.aimsonnx import stylize_model_comparison
        stylized_compare = stylize_model_comparison(comp_dict_out=compare_dict)
        return(stylized_compare)

    def inspect_y_test(self):
        from aimodelshare.aimsonnx import inspect_y_test as inspect
        data = inspect(apiurl = self.playground_url)
        return data 
    
    def get_leaderboard(self, verbose=3, columns=None):
        from aimodelshare.leaderboard import get_leaderboard as get_lead
        data = get_lead(verbose=verbose,
                 columns=columns, 
                 apiurl = self.playground_url)
        return data
    
    def stylize_leaderboard(self, leaderboard):
        from aimodelshare.leaderboard import stylize_leaderboard as stylize_lead
        stylized_leaderboard = stylize_lead(leaderboard = leaderboard)
        return stylized_leaderboard
    
    def update_access_list(self, email_list=[],update_type="Replace_list"): 
        from aimodelshare.generatemodelapi import update_access_list as update_list
        update = update_list(apiurl = self.playground_url, email_list=email_list,update_type=update_type)
        return update







class Data: 
    def __init__(self, data_type, playground_url=None):
        self.data_type = data_type
        self.playground_url = playground_url 
    
    def __str__(self):
        return f"This is a description of the Data class."

    def share_dataset(self, data_directory="folder_file_path", classification="default", private="FALSE"): 
        from aimodelshare.data_sharing.share_data import share_dataset as share
        response = share(data_directory=data_directory, classification=classification, private=private)
        return response
    
    def download_data(self, repository):
        from aimodelshare.data_sharing.download_data import download_data as download
        datadownload = download(repository)
        return datadownload




#import docker
import os
import shutil
import importlib.resources as pkg_resources
#import importlib_resources as pkg_resources
from . import data_sharing_templates
from string import Template
import zipfile
import time
import json
import boto3
import tempfile
import requests
import uuid

def create_docker_folder_local(dataset_dir, dataset_name, python_version):

    tmp_dataset_dir = tempfile.gettempdir() + '/' + '/'.join(['tmp_dataset_dir', dataset_name])

    tmp_dataset = tempfile.gettempdir() + '/' + 'tmp_dataset_dir'

    os.mkdir(tmp_dataset)

    shutil.copytree(dataset_dir, tmp_dataset_dir)

    #data = pkg_resources.read_text(data_sharing_templates, 'Dockerfile.txt')
    with open(os.path.join('data_sharing_templates', 'Dockerfile.txt'), 'r') as file:
        data = file.read()

    template = Template(data)
    newdata = template.substitute(
        python_version=python_version)
    with open(os.path.join(tmp_dataset, 'Dockerfile'), 'w') as file:
        file.write(newdata)

def create_docker_folder_codebuild(dataset_dir, dataset_name, template_folder, region, registry_uri, repository, dataset_tag, python_version):

    tmp_dataset_dir = tempfile.gettempdir() + '/' + '/'.join(['tmp_dataset_dir', dataset_name])

    tmp_dataset = tempfile.gettempdir() + '/' + 'tmp_dataset_dir'

    os.mkdir(tmp_dataset)

    """
    if not os.path.exists(tmp_dataset):
        os.makedirs(tmp_dataset)
    else:
        shutil.rmtree(tmp_dataset)
        os.makedirs(tmp_dataset)
    """

    shutil.copytree(dataset_dir, tmp_dataset_dir)

    os.mkdir(template_folder)

    data = pkg_resources.read_text(data_sharing_templates, 'Dockerfile.txt')
    #with open(os.path.join('data_sharing_templates', 'Dockerfile.txt'), 'r') as file:
    #    data = file.read()

    template = Template(data)
    newdata = template.substitute(
        python_version=python_version)
    with open(os.path.join(template_folder, 'Dockerfile'), 'w') as file:
        file.write(newdata)
    
    data = pkg_resources.read_text(data_sharing_templates, 'buildspec.txt')
    #with open(os.path.join('data_sharing_templates', 'buildspec.txt'), 'r') as file:
    #    data = file.read()

    template = Template(data)
    newdata = template.substitute(
        aws_access_key_id=os.environ['AWS_ACCESS_KEY_ID'],
        aws_secret_access_key=os.environ['AWS_SECRET_ACCESS_KEY'],
        region=region,
        registry_uri=registry_uri,
        repository=repository,
        dataset_tag=dataset_tag)
    with open(os.path.join(template_folder, 'buildspec.yml'), 'w') as file:
        file.write(newdata)

    response = shutil.copytree(tmp_dataset, '/'.join([template_folder, 'tmp_dataset_dir']))

    def get_all_file_paths(directory):
        file_paths = []
        for root, directories, files in os.walk(directory):
            for filename in files:
                filepath = os.path.join(root, filename)
                file_paths.append(filepath)
        return file_paths

    file_paths = get_all_file_paths(template_folder)

    template_folder_len = len(template_folder)

    with zipfile.ZipFile(''.join([template_folder, '.zip']),'w') as zip:
        for file in file_paths:
            try:
                zip.write(file, file[template_folder_len:])
            except:
                pass

    shutil.rmtree(tmp_dataset)
    
    shutil.rmtree(template_folder)

#def share_data_local(dataset_dir, tag='latest', python_version='3.8'):

    #create_docker_folder_local(dataset_dir)

    #client = docker.from_env()

    #client.images.build(path='./tmp_dataset_folder', tag=tag)

    #shutil.rmtree('tmp_dataset_dir')

    # send to ecr
    
    # client.images.

    # client.push(repository,

    # ecr_client.create_repository

    # docker tag aimodelshare-base-image:latest 517169013426.dkr.ecr.us-east-1.amazonaws.com/aimodelshare-base-image:latest
    # aws ecr-public get-login-password --region us-east-1 | docker login --username AWS --password-stdin public.ecr.aws
    # docker push public.ecr.aws/y2e2a1d6/aimodelshare-image-classification-public

def share_data_codebuild(account_id, region, dataset_dir, dataset_tag='latest', python_version='3.8'):

    print('Uploading your data. Please wait for a confirmation message.')
    
    session = boto3.session.Session(aws_access_key_id=os.environ.get("AWS_ACCESS_KEY_ID"),
                                    aws_secret_access_key = os.environ.get("AWS_SECRET_ACCESS_KEY"), 
                                    region_name=os.environ.get("AWS_REGION"))

    flag = 0

    dataset_name = dataset_dir.replace(" ", "_")

    repository=dataset_name+'-repository'

    template_folder=tempfile.gettempdir() + '/' + dataset_name+'_'+dataset_tag

    codebuild_role_name=dataset_name+'-codebuild-role'
    codebuild_policies_name=dataset_name+'-codebuild-policies'

    codebuild_dataset_name=dataset_name+'-upload'

    s3_client = session.resource('s3', region_name=region)

    bucket_name = "aimodelshare"+str(account_id)+"sharedata"
    
    try:
        s3_client.create_bucket(
            Bucket=bucket_name,
            CreateBucketConfiguration = {
                'LocationConstraint': region
            }
        )
    except:
        s3_client.create_bucket(
            Bucket=bucket_name
        )
    
    s3_resource = session.resource('s3', region_name=region)

    bucket_versioning = s3_resource.BucketVersioning(bucket_name)
    response = bucket_versioning.enable()

    # ecr = session.client('ecr')

    ecr = session.client('ecr-public')

    registry_uri = ecr.describe_registries()['registries'][0]['registryUri']

    try:
        response = ecr.create_repository(
            repositoryName=repository
        )
    except:
        pass

    create_docker_folder_codebuild(dataset_dir, dataset_name, template_folder, region, registry_uri, repository, dataset_tag, python_version)

    iam = session.client('iam')
    
    codebuild_trust_relationship = json.loads(pkg_resources.read_text(data_sharing_templates, 'codebuild_trust_relationship.txt'))
    #with open(os.path.join('data_sharing_templates', 'codebuild_trust_relationship.txt'), 'r') as file:
    #    codebuild_trust_relationship = json.load(file)

    try:
        response = iam.create_role(
            RoleName=codebuild_role_name,
            AssumeRolePolicyDocument=json.dumps(codebuild_trust_relationship)
        )
    except:
        flag = 1
        pass

    codebuild_policies = json.loads(pkg_resources.read_text(data_sharing_templates, 'codebuild_policies.txt'))
    #with open(os.path.join('data_sharing_templates', 'codebuild_policies.txt'), 'r') as file:
    #    codebuild_policies = json.load(file)

    try:
        response = iam.create_policy(
            PolicyName=codebuild_policies_name,
            PolicyDocument=json.dumps(codebuild_policies),
        )
    except:
        flag = 1
        pass

    response = iam.attach_role_policy(
        RoleName=codebuild_role_name,
        PolicyArn=''.join(['arn:aws:iam::', account_id, ':policy/', codebuild_policies_name])
    )

    s3_client = session.client('s3')
    s3_client.upload_file(''.join([template_folder, '.zip']),
                          bucket_name,
                          ''.join([dataset_name+'_'+dataset_tag, '.zip']))

    if(flag==0):
        time.sleep(15)

    codebuild = session.client('codebuild')
    
    try:
        response = codebuild.create_project(
            name=codebuild_dataset_name,
            source={
                'type': 'S3',
                'location': bucket_name + '/' + dataset_name+'_'+dataset_tag + '.zip'
            },
            artifacts={
                'type': 'NO_ARTIFACTS',
            },
            environment={
                'type': 'LINUX_CONTAINER',
                'image': 'aws/codebuild/standard:5.0',
                'computeType': 'BUILD_GENERAL1_SMALL',
                'privilegedMode': True
            },
            serviceRole=codebuild_role_name
        )
    except:
        response = codebuild.delete_project(
            name=codebuild_dataset_name
        )
        response = codebuild.create_project(
            name=codebuild_dataset_name,
            source={
                'type': 'S3',
                'location': bucket_name + '/' + dataset_name+'_'+dataset_tag + '.zip'
            },
            artifacts={
                'type': 'NO_ARTIFACTS',
            },
            environment={
                'type': 'LINUX_CONTAINER',
                'image': 'aws/codebuild/standard:5.0',
                'computeType': 'BUILD_GENERAL1_SMALL',
                'privilegedMode': True
            },
            serviceRole=codebuild_role_name
        )

    response = codebuild.start_build(
        projectName=codebuild_dataset_name
    )

    os.remove(template_folder+'.zip')

    return {"ecr_uri":registry_uri + '/' + repository + ':' + dataset_tag}

def share_dataset(data_directory="folder_file_path",classification="default", private="FALSE"):
    data_directory=str(data_directory).lower()
    aishare_datasetname = input("Enter dataset name:")
    aishare_datadescription = input(
        "Enter data description (i.e.- filenames denoting training and test data, file types, and any subfolders where files are stored):")
    aishare_datatags = input(
        "Enter tags to help users find your data (i.e.- flower dataset, image, supervised learning, classification")   
    datalicense=input("Insert license (Optional): ")
    datacitation=input("Insert citation (Optional): ")
    modelplaygroundurl=input("Insert AI Model Share model playground url (Optional): ")
    problemdomain=input("Enter a number signifying your dataset problem domain or data type: 1 = Image 2 = Video 3 = Text 4 = Tabular 5 = Neural Style Transfer 6 = Object Detection 7 = Other \n")

    optiondict={"1":"Image", "2":"Video","3":"Text","4":"Tabular", "5": "Audio","6":"Neural Style Transfer", "7": "Object Detection", "8":"Other"}
    problemdomainfinal=optiondict.get(problemdomain,"Other")


    user_session = boto3.session.Session(aws_access_key_id=os.environ.get("AWS_ACCESS_KEY_ID"),
                                          aws_secret_access_key = os.environ.get("AWS_SECRET_ACCESS_KEY"), 
                                         region_name=os.environ.get("AWS_REGION"))
    account_number = user_session.client(
        'sts').get_caller_identity().get('Account')

    datauri=share_data_codebuild(account_number,os.environ.get("AWS_REGION"),data_directory)


    #TODO: Replace redis data code with new api post call, see helper code below from competition api
    #TODO: add "dataset:id" as models are ingested on the backend
    bodydata = {"dataowner": os.environ.get("username"),  # change this to first and last name
                "dataname":aishare_datasetname ,                
                'datadescription':aishare_datadescription,
                'datatags':aishare_datatags,
                'dataecruri':datauri['ecr_uri'],
                'datalicense':datalicense,
                'datacitation':datacitation,
                'classification':classification,
                "modelplaygroundurl": modelplaygroundurl,
                "Private": private,
                "delete": "FALSE",
                "problemdomain":problemdomainfinal}

    # datasets api
    headers_with_authentication = {'Content-Type': 'application/json', 'authorizationToken': os.environ.get("JWT_AUTHORIZATION_TOKEN"), 'Access-Control-Allow-Headers':
                                   'Content-Type,X-Amz-Date,authorizationToken,Access-Control-Allow-Origin,X-Api-Key,X-Amz-Security-Token,Authorization', 'Access-Control-Allow-Origin': '*'}
    # modeltoapi lambda function invoked through below url to return new prediction api in response
    response=requests.post("https://jyz9nn0joe.execute-api.us-east-1.amazonaws.com/dev/modeldata",
                  json=bodydata, headers=headers_with_authentication)
    return "Your dataset has been shared to modelshare.org."



import os
import boto3
import json
import onnx
import numpy as np
import pandas as pd
import requests 
import json
import ast
import tempfile
import tensorflow as tf

from datetime import datetime

from aimodelshare.aws import run_function_on_lambda, get_token, get_aws_token, get_aws_client

from aimodelshare.aimsonnx import _get_leaderboard_data, inspect_model, _get_metadata, _model_summary, model_from_string


def _get_file_list(client, bucket, model_id):
    #  Reading file list {{{
    try:
        objects = client["client"].list_objects(Bucket=bucket, Prefix=model_id + "/")
    except Exception as err:
        return None, err

    file_list = []
    if "Contents" in objects:
        for key in objects["Contents"]:
            file_list.append(key["Key"].split("/")[1])
    #  }}}

    return file_list, None


def _delete_s3_object(client, bucket, model_id, filename):
    deletionobject = client["resource"].Object(bucket, model_id + "/" + filename)
    deletionobject.delete()

def _get_predictionmodel_key(unique_model_id,file_extension):
    if file_extension==".pkl":
        file_key = unique_model_id + "/runtime_model" + file_extension
        versionfile_key = unique_model_id + "/predictionmodel_1" + file_extension
    else:
        file_key = unique_model_id + "/runtime_model" + file_extension
        versionfile_key = unique_model_id + "/predictionmodel_1" + file_extension
    return file_key,versionfile_key


def _upload_onnx_model(modelpath, client, bucket, model_id, model_version):
    # Check the model {{{
    if not os.path.exists(modelpath):
        raise FileNotFoundError(f"The model file at {modelpath} does not exist")

    file_name = os.path.basename(modelpath)
    file_name, file_ext = os.path.splitext(file_name)

    assert (
        file_ext == ".onnx"
    ), "modelshareai api only supports .onnx models at the moment"
    # }}}

    # Upload the model {{{
    try:
        client["client"].upload_file(
            modelpath, bucket, model_id + "/onnx_model_mostrecent.onnx"
        )
        client["client"].upload_file(
            modelpath,
            bucket,
            model_id + "/onnx_model_v" + str(model_version) + file_ext,
        )
    except Exception as err:
        return err
    # }}}

def _upload_native_model(modelpath, client, bucket, model_id, model_version):
    # Check the model {{{
    if not os.path.exists(modelpath):
        raise FileNotFoundError(f"The model file at {modelpath} does not exist")

    file_name = os.path.basename(modelpath)
    file_name, file_ext = os.path.splitext(file_name)

    assert (
        file_ext == ".onnx"
    ), "modelshareai api only supports .onnx models at the moment"
    # }}}

    # Upload the model {{{
    try:
        client["client"].upload_file(
            modelpath, bucket, model_id + "/onnx_model_mostrecent.onnx"
        )
        client["client"].upload_file(
            modelpath,
            bucket,
            model_id + "/onnx_model_v" + str(model_version) + file_ext,
        )
    except Exception as err:
        return err
    # }}}


def _upload_preprocessor(preprocessor, client, bucket, model_id, model_version):

  try:

    
    # Check the preprocessor {{{
    if not os.path.exists(preprocessor):
        raise FileNotFoundError(
            f"The preprocessor file at {preprocessor} does not exist"
        )

    
    file_name = os.path.basename(preprocessor)
    file_name, file_ext = os.path.splitext(file_name)
    
    from zipfile import ZipFile
    dir_zip = preprocessor

    #zipObj = ZipFile(os.path.join("./preprocessor.zip"), 'a')
    #/Users/aishwarya/Downloads/aimodelshare-master
    client["client"].upload_file(dir_zip, bucket, model_id + "/preprocessor_v" + str(model_version)+ ".zip")
  except Exception as e:
    print(e)


def _update_leaderboard(
    modelpath, eval_metrics, client, bucket, model_id, model_version
):
    # Loading the model and its metadata {{{
    if not os.path.exists(modelpath):
        raise FileNotFoundError(f"The model file at {modelpath} does not exist")

    model = onnx.load(modelpath)
    metadata = _get_leaderboard_data(model, eval_metrics)
    # }}}

    # Adding extra details to metadata {{{
    metadata["username"] = os.environ.get("username")
    metadata["timestamp"] = str(datetime.now())
    metadata["version"] = model_version
    # }}}
    
    #TODO: send above data in post call to /eval and update master table on back end rather than downloading locally.
    #Either way something is breaking and the s3 version should still work right??
    # Read existing table {{{
    try:
        leaderboard = client["client"].get_object(
            Bucket=bucket, Key=model_id + "/model_eval_data_mastertable.csv"



        )
        leaderboard = pd.read_csv(leaderboard["Body"], sep="\t")
        columns = leaderboard.columns

    except client["client"].exceptions.NoSuchKey:
        # Create leaderboard if not exists
        # FIXME: Find a better way to get columns
        columns = list(metadata.keys())
        leaderboard = pd.DataFrame(columns=columns)

    except Exception as err:
        raise err
    # }}}

    # Update the leaderboard {{{
    metadata = {col: metadata.get(col, None) for col in columns}
    leaderboard = leaderboard.append(metadata, ignore_index=True, sort=False)

    leaderboard_csv = leaderboard.to_csv(index=False, sep="\t")
    metadata.pop("model_config", "pop worked")

    try:
        s3_object = client["resource"].Object(
            bucket, model_id + "/model_eval_data_mastertable.csv"
        )
        s3_object.put(Body=leaderboard_csv)
        return metadata
    except Exception as err:
        return err
    # }}}

def _update_leaderboard_public(
    modelpath, eval_metrics,s3_presigned_dict):
    # Loading the model and its metadata {{{
    if not os.path.exists(modelpath):
        raise FileNotFoundError(f"The model file at {modelpath} does not exist")

    model_versions = [os.path.splitext(f)[0].split("_")[-1][1:] for f in s3_presigned_dict['put'].keys()]
    
    model_versions = filter(lambda v: v.isnumeric(), model_versions)
    model_versions = list(map(int, model_versions))
    model_version=model_versions[0]

    model_version

    model = onnx.load(modelpath)
    metadata = _get_leaderboard_data(model, eval_metrics)
    # }}}

    # Adding extra details to metadata {{{
    metadata["username"] = os.environ.get("username")
    metadata["timestamp"] = str(datetime.now())
    metadata["version"] = model_version
    # }}}
    import tempfile
    temp=tempfile.mkdtemp()
    #TODO: send above data in post call to /eval and update master table on back end rather than downloading locally.
    #Either way something is breaking and the s3 version should still work right??
    # Read existing table {{{
    try:
        import wget


        #Get leaderboard
        leaderboardfilename = wget.download(s3_presigned_dict['get']['model_eval_data_mastertable.csv'], out=temp+"/"+'model_eval_data_mastertable.csv')
        import pandas as pd
        leaderboard=pd.read_csv(temp+"/"+'model_eval_data_mastertable.csv', sep="\t")

        columns = leaderboard.columns
        
    except:
        # Create leaderboard if not exists
        # FIXME: Find a better way to get columns
        import pandas as pd
        columns = list(metadata.keys())
        leaderboard = pd.DataFrame(columns=columns)

    # }}}

    # Update the leaderboard {{{
    metadata = {col: metadata.get(col, None) for col in columns}
    leaderboard = leaderboard.append(metadata, ignore_index=True, sort=False)
    
    leaderboard_csv = leaderboard.to_csv(temp+"/"+'model_eval_data_mastertable.csv',index=False, sep="\t")
    metadata.pop("model_config", "pop worked")

    try:

      putfilekeys=list(s3_presigned_dict['put'].keys())
      modelputfiles = [s for s in putfilekeys if str("csv") in s]

      fileputlistofdicts=[]
      for i in modelputfiles:
        filedownload_dict=ast.literal_eval(s3_presigned_dict ['put'][i])
        fileputlistofdicts.append(filedownload_dict)

      with open(temp+"/"+'model_eval_data_mastertable.csv', 'rb') as f:
        files = {'file': (temp+"/"+'model_eval_data_mastertable.csv', f)}
        http_response = requests.post(fileputlistofdicts[0]['url'], data=fileputlistofdicts[0]['fields'], files=files)
        return metadata
    except Exception as err:
        return err
 

def upload_model_dict(modelpath, s3_presigned_dict, bucket, model_id, model_version):
    import wget
    import json
    import tempfile
    temp=tempfile.mkdtemp()
    # get model summary from onnx
    onnx_model = onnx.load(modelpath)
    meta_dict = _get_metadata(onnx_model)

    if meta_dict['ml_framework'] == 'keras':

        inspect_pd = _model_summary(meta_dict)
        
    elif meta_dict['ml_framework'] in ['sklearn', 'xgboost']:

        model_config = meta_dict["model_config"]
        model_config = ast.literal_eval(model_config)

        model_class = model_from_string(meta_dict['model_type'])
        default = model_class()
        default_config = default.get_params()

        inspect_pd = pd.DataFrame({'param_name': model_config.keys(),
                                   'default_value': default_config.values(),
                                   'param_value': model_config.values()})
   
    try:
        #Get inspect json
        inspectdatafilename = wget.download(s3_presigned_dict['get']['inspect_pd.json'], out=temp+"/"+'inspect_pd.json')
        
        with open(temp+"/"+'inspect_pd.json') as f:
            model_dict  = json.load(f)
    except: 
      model_dict = {}

    model_dict[str(model_version)] = {'ml_framework': meta_dict['ml_framework'],
                                      'model_type': meta_dict['model_type'],
                                      'model_dict': inspect_pd.to_dict()}

    with open(temp+"/"+'inspect_pd.json', 'w') as outfile:
        json.dump(model_dict, outfile)

    try:

      putfilekeys=list(s3_presigned_dict['put'].keys())
      modelputfiles = [s for s in putfilekeys if str("json") in s]

      fileputlistofdicts=[]
      for i in modelputfiles:
        filedownload_dict=ast.literal_eval(s3_presigned_dict ['put'][i])
        fileputlistofdicts.append(filedownload_dict)

      with open(temp+"/"+'inspect_pd.json', 'rb') as f:
        files = {'file': (temp+"/"+'inspect_pd.json', f)}
        http_response = requests.post(fileputlistofdicts[0]['url'], data=fileputlistofdicts[0]['fields'], files=files)
    except:
      pass
    return 1


def submit_model(
    model_filepath=None,
    apiurl=None,
    prediction_submission=None,
    preprocessor=None
    ):
    """
    Submits model/preprocessor to machine learning competition using live prediction API url generated by AI Modelshare library
    The submitted model gets evaluated and compared with all existing models and a leaderboard can be generated 
    ---------------
    Parameters:
    modelpath:  string ends with '.onnx'
                value - Absolute path to model file [REQUIRED] to be set by the user
                .onnx is the only accepted model file extension
                "example_model.onnx" filename for file in directory.
                "/User/xyz/model/example_model.onnx" absolute path to model file from local directory
    apiurl :    string 
                value - url to the live prediction REST API generated for the user's model 
                "https://example.execute-api.us-east-1.amazonaws.com/prod/m"
    prediction_submission:   one hot encoded y_pred
                    value - predictions for test data
                    [REQUIRED] for evaluation metriicts of the submitted model
    preprocessor:   string,default=None
                    value - absolute path to preprocessor file 
                    [REQUIRED] to be set by the user
                    "./preprocessor.zip" 
                    searches for an exported zip preprocessor file in the current directory
                    file is generated from preprocessor module using export_preprocessor function from the AI Modelshare library 
                   
    -----------------
    Returns
    response:   Model version if the model is submitted sucessfully
                error  if there is any error while submitting models
    
    """

    import os
    from aimodelshare.aws import get_aws_token
    from aimodelshare.modeluser import get_jwt_token, create_user_getkeyandpassword

    # Confirm that creds are loaded, print warning if not
    if all(["username" in os.environ, 
            "password" in os.environ]):
        pass
    else:
        return print("'Submit Model' unsuccessful. Please provide username and password using set_credentials() function.")

    apiurl=apiurl.replace('"','')

    # Get bucket and model_id for user {{{
    response, error = run_function_on_lambda(
        apiurl, **{"delete": "FALSE", "versionupdateget": "TRUE"}
    )
    if error is not None:
        raise error

    _, bucket, model_id = json.loads(response.content.decode("utf-8"))
    # }}}

    #begin replacing code here
    #add call to eval lambda here to retrieve presigned urls and eval metrics
    if prediction_submission is not None:
        if type(prediction_submission) is not list:
            prediction_submission=prediction_submission.tolist()
        else: 
            pass

        if all(isinstance(x, (np.float64)) for x in prediction_submission):
              prediction_submission = [float(i) for i in prediction_submission]
        else: 
            pass

    try:

        post_dict = {"y_pred": prediction_submission,
                "return_eval": "True",
                "return_y": "False"}

        headers = { 'Content-Type':'application/json', 'authorizationToken': os.environ.get("AWS_TOKEN"), } 
        apiurl_eval=apiurl[:-1]+"eval"
        prediction = requests.post(apiurl_eval,headers=headers,data=json.dumps(post_dict)) 

        eval_metrics=json.loads(prediction.text)
    except:
        pass

    if all([isinstance(eval_metrics, dict),"message" not in eval_metrics]):
        pass        
    else:
        if all([isinstance(eval_metrics, list)]):
            return print(eval_metrics[0])
        else:
            return print('Unauthorized user: You do not have access to submit models to, or request data from, this competition.')
    

    if all(value == None for value in eval_metrics.values()):
        return print("Failed to calculate evaluation metrics. Please check the format of the submitted predictions.")

    s3_presigned_dict = {key:val for key, val in eval_metrics.items() if key != 'eval'}
    eval_metrics = {key:val for key, val in eval_metrics.items() if key != 'get'}
    eval_metrics = {key:val for key, val in eval_metrics.items() if key != 'put'}
    if eval_metrics.get("eval","empty")=="empty":
      pass
    else:
      eval_metrics=eval_metrics['eval']
    
    #upload preprocessor (1s for small upload vs 21 for 306 mbs)
    putfilekeys=list(s3_presigned_dict['put'].keys())
    modelputfiles = [s for s in putfilekeys if str("zip") in s]

    fileputlistofdicts=[]
    for i in modelputfiles:
      filedownload_dict=ast.literal_eval(s3_presigned_dict ['put'][i])
      fileputlistofdicts.append(filedownload_dict)


    with open(preprocessor, 'rb') as f:
      files = {'file': (preprocessor, f)}
      http_response = requests.post(fileputlistofdicts[0]['url'], data=fileputlistofdicts[0]['fields'], files=files)

    putfilekeys=list(s3_presigned_dict['put'].keys())
    modelputfiles = [s for s in putfilekeys if str("onnx") in s]

    fileputlistofdicts=[]
    for i in modelputfiles:
      filedownload_dict=ast.literal_eval(s3_presigned_dict ['put'][i])
      fileputlistofdicts.append(filedownload_dict)


    with open(model_filepath, 'rb') as f:
      files = {'file': (model_filepath, f)}
      http_response = requests.post(fileputlistofdicts[1]['url'], data=fileputlistofdicts[1]['fields'], files=files)
      

    # Upload model metrics and metadata {{{
    modelleaderboarddata = _update_leaderboard_public(
        model_filepath, eval_metrics,s3_presigned_dict
    )

    model_versions = [os.path.splitext(f)[0].split("_")[-1][1:] for f in s3_presigned_dict['put'].keys()]

    model_versions = filter(lambda v: v.isnumeric(), model_versions)
    model_versions = list(map(int, model_versions))
    model_version=model_versions[0]

    upload_model_dict(model_filepath, s3_presigned_dict, bucket, model_id, model_version)

    modelpath=model_filepath

    def dict_clean(items):
      result = {}
      for key, value in items:
          if value is None:
              value = '0'
          result[key] = value
      return result

    if isinstance(modelleaderboarddata, Exception):
      raise err
    else:
      dict_str = json.dumps(modelleaderboarddata)
    #convert None type values to string
      modelleaderboarddata_cleaned = json.loads(dict_str, object_pairs_hook=dict_clean)

    # Update model version and sample data {{{
    #data_types = None
    #data_columns = None
    #if sample_data is not None and isinstance(sample_data, pd.DataFrame):
    #    data_types = list(sample_data.dtypes.values.astype(str))
    #    data_columns = list(sample_data.columns)

    #kwargs = {
    #    "delete": "FALSE",
    #    "versionupdateget": "FALSE",
    #    "versionupdateput": "TRUE",
    #    "version": model_version,
    #    "input_feature_dtypes": data_types,
    #    "input_feature_names": data_columns,
    #}
    #response, error = run_function_on_lambda(apiurl, aws_token, **kwargs)
    #if error is not None:
    #    raise error
    # }}}
    modelsubmissiontags=input("Insert search tags to help users find your model (optional): ")
    modelsubmissiondescription=input("Provide any useful notes about your model (optional): ")

    #Update competition data
    bodydata = {"apiurl": apiurl,
                "submissions": model_version,
                  "contributoruniquenames":os.environ.get('username'),
                "versionupdateputsubmit":"TRUE"
                                }

    # Get the response
    headers_with_authentication = {'Content-Type': 'application/json', 'authorizationToken': os.environ.get("AWS_TOKEN"), 'Access-Control-Allow-Headers':
                                    'Content-Type,X-Amz-Date,authorizationToken,Access-Control-Allow-Origin,X-Api-Key,X-Amz-Security-Token,Authorization', 'Access-Control-Allow-Origin': '*'}
    # competitiondata lambda function invoked through below url to update model submissions and contributors
    requests.post("https://o35jwfakca.execute-api.us-east-1.amazonaws.com/dev/modeldata",
                  json=bodydata, headers=headers_with_authentication)


    # get model summary from onnx
    onnx_model = onnx.load(modelpath)
    meta_dict = _get_metadata(onnx_model)

    if meta_dict['ml_framework'] == 'keras':

        inspect_pd = _model_summary(meta_dict)
        
    elif meta_dict['ml_framework'] in ['sklearn', 'xgboost']:

        model_config = meta_dict["model_config"]
        model_config = ast.literal_eval(model_config)

        model_class = model_from_string(meta_dict['model_type'])
        default = model_class()
        default_config = default.get_params()

        inspect_pd = pd.DataFrame({'param_name': model_config.keys(),
                                   'default_value': default_config.values(),
                                   'param_value': model_config.values()})
   
    keys_to_extract = [ "accuracy", "f1_score", "precision", "recall", "mse", "rmse", "mae", "r2"]

    eval_metrics_subset = {key: eval_metrics[key] for key in keys_to_extract}

    eval_metrics_subset_nonulls = {key: value for key, value in eval_metrics_subset.items() if isinstance(value, float)}

                              
    #Update model architecture data
    bodydatamodels = {
                "apiurl": apiurl,
                "modelsummary":json.dumps(inspect_pd.to_json()),
                "Private":"FALSE",
                "modelsubmissiondescription": modelsubmissiondescription,
                "modelsubmissiontags":modelsubmissiontags,
                 "eval_metrics":json.dumps(eval_metrics_subset_nonulls)}

    bodydatamodels.update(modelleaderboarddata_cleaned)
    d = bodydatamodels


    keys_values = d.items()


    bodydatamodels_allstrings = {str(key): str(value) for key, value in keys_values}



    # Get the response
    headers_with_authentication = {'Content-Type': 'application/json', 'authorizationToken': os.environ.get("AWS_TOKEN"), 'Access-Control-Allow-Headers':
                                    'Content-Type,X-Amz-Date,authorizationToken,Access-Control-Allow-Origin,X-Api-Key,X-Amz-Security-Token,Authorization', 'Access-Control-Allow-Origin': '*'}
    # competitiondata lambda function invoked through below url to update model submissions and contributors
    response=requests.post("https://eeqq8zuo9j.execute-api.us-east-1.amazonaws.com/dev/modeldata",
                  json=bodydatamodels_allstrings, headers=headers_with_authentication)

    if str(response.status_code)=="200":
        code_comp_result="To submit code used to create this model or to view current leaderboard navigate to Model Playground: \n\n https://www.modelshare.org/detail/model:"+response.text.split(":")[1]  
    else:
        code_comp_result="" #TODO: reponse 403 indicates that user needs to reset credentials.  Need to add a creds check to top of function.


    return print("\nYour model has been submitted as model version "+str(model_version)+ "\n\n"+code_comp_result)

def update_runtime_model(apiurl, model_version=None):
    """
    apiurl: string of API URL that the user wishes to edit
    new_model_version: string of model version number (from leaderboard) to replace original model 
    """
    # Confirm that creds are loaded, print warning if not
    if all(["AWS_ACCESS_KEY_ID" in os.environ, 
            "AWS_SECRET_ACCESS_KEY" in os.environ,
            "AWS_REGION" in os.environ,
           "username" in os.environ, 
           "password" in os.environ]):
        pass
    else:
        return print("'Update Runtime Model' unsuccessful. Please provide credentials with set_credentials().")

    # Create user session
    aws_client=get_aws_client(aws_key=os.environ.get('AWS_ACCESS_KEY_ID'), 
                              aws_secret=os.environ.get('AWS_SECRET_ACCESS_KEY'), 
                              aws_region=os.environ.get('AWS_REGION'))
    
    user_sess = boto3.session.Session(aws_access_key_id=os.environ.get('AWS_ACCESS_KEY_ID'), 
                                      aws_secret_access_key=os.environ.get('AWS_SECRET_ACCESS_KEY'), 
                                      region_name=os.environ.get('AWS_REGION'))
    
    s3 = user_sess.resource('s3')
    model_version=str(model_version)
    # Get bucket and model_id for user based on apiurl {{{
    response, error = run_function_on_lambda(
        apiurl, **{"delete": "FALSE", "versionupdateget": "TRUE"}
    )
    if error is not None:
        raise error

    _, api_bucket, model_id = json.loads(response.content.decode("utf-8"))
    # }}}

    try:
        leaderboard = aws_client["client"].get_object(
            Bucket=api_bucket, Key=model_id + "/model_eval_data_mastertable.csv"



        )
        leaderboard = pd.read_csv(leaderboard["Body"], sep="\t")
        columns = leaderboard.columns
        metric_names=["accuracy","f1_score","precision","recall","r2","mse","mae"]
        leaderboardversion=leaderboard[leaderboard['version']==int(model_version)]
        leaderboardversion=leaderboardversion.dropna(axis=1)
        metric_names_subset=list(set(metric_names).intersection(leaderboardversion.columns))
        leaderboardversiondict=leaderboardversion.loc[:,metric_names_subset].to_dict('records')[0]
    except Exception as err:
        raise err

    # Get file list for current bucket {{{
    model_files, err = _get_file_list(aws_client, api_bucket, model_id)
    if err is not None:
        raise err
    # }}}

    # extract subfolder objects specific to the model id
    folder = s3.meta.client.list_objects(Bucket=api_bucket, Prefix=model_id+"/")
    bucket = s3.Bucket(api_bucket)
    file_list = [file['Key'] for file in folder['Contents']]
    s3 = boto3.resource('s3')
    model_source_key = model_id+"/onnx_model_v"+str(model_version)+".onnx"
    preprocesor_source_key = model_id+"/preprocessor_v"+str(model_version)+".zip"
    model_copy_source = {
          'Bucket': api_bucket,
          'Key': model_source_key
        }
    preprocessor_copy_source = {
          'Bucket': api_bucket,
          'Key': preprocesor_source_key
      }
    # Sending correct model metrics to front end 
    bodydatamodelmetrics={"apiurl":apiurl,
                          "versionupdateput":"TRUE",
                          "verified_metrics":"TRUE",
                          "eval_metrics":json.dumps(leaderboardversiondict)}
 
    headers = { 'Content-Type':'application/json', 'authorizationToken': os.environ.get("AWS_TOKEN"), } 
    prediction = requests.post("https://bhrdesksak.execute-api.us-east-1.amazonaws.com/dev/modeldata",headers=headers,data=json.dumps(bodydatamodelmetrics)) 

    # overwrite runtime_model.onnx file & runtime_preprocessor.zip files: 
    if (model_source_key in file_list) & (preprocesor_source_key in file_list):
        response = bucket.copy(model_copy_source, model_id+"/"+'runtime_model.onnx')
        response = bucket.copy(preprocessor_copy_source, model_id+"/"+'runtime_preprocessor.zip')
        return print('Runtime model & preprocessor for api: '+apiurl+" updated to model version "+model_version+".\n\nModel metrics are now updated and verified for this model playground.")
    else:
        # the file resource to be the new runtime_model is not available
        return 'New Runtime Model version ' + model_version + ' not found.'
    

def _extract_model_metadata(model, eval_metrics=None):
    # Getting the model metadata {{{
    graph = model.graph

    if eval_metrics is not None:
        metadata = eval_metrics
    else:
        metadata = dict()

    metadata["num_nodes"] = len(graph.node)
    metadata["depth_test"] = len(graph.initializer)
    metadata["num_params"] = sum(np.product(node.dims) for node in graph.initializer)

    # layers = ""
    # for node in graph.node:
    #     # consider type and get node attributes (??)
    #     layers += (
    #         node.op_type
    #         + "x".join(str(d.ints) for d in node.attribute if hasattr(d, 'ints'))
    #     )
    metadata["layers"] = "; ".join(node.op_type for node in graph.node)

    inputs = ""
    for inp in graph.input:
        dims = []
        for d in inp.type.tensor_type.shape.dim:
            if d.dim_param != "":
                dims.append(d.dim_param)
            else:
                dims.append(str(d.dim_value))

        metadata["input_shape"] = dims
        inputs += f"{inp.name} ({'x'.join(dims)})"
    metadata["inputs"] = inputs

    outputs = ""
    for out in graph.output:
        dims = []
        for d in out.type.tensor_type.shape.dim:
            if d.dim_param != "":
                dims.append(d.dim_param)
            else:
                dims.append(str(d.dim_value))

        outputs += f"{out.name} ({'x'.join(dims)})"
    metadata["outputs"] = outputs
    # }}}

    return metadata

__all__ = [
    submit_model,
    _extract_model_metadata,
    update_runtime_model
]




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='aimodelshare', #TODO:update
    version='0.0.32',        #TODO:update
    author="Michael Parrott",
    author_email="mikedparrott@modelshare.org",
    description="Deploy locally saved machine learning models to a live rest API and web-dashboard.  Share it with the world via modelshare.org",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://www.modelshare.org",
    packages=setuptools.find_packages(),
    install_requires=["boto3==1.18.2", "botocore==1.21.2", "scikit-learn==0.24.2","onnx>=1.9.0","onnxconverter-common>=1.7.0", 
    "keras2onnx>=1.7.0","tensorflow>=2.5.0","tf2onnx","skl2onnx>=1.8.0","onnxruntime>=1.7.0","urllib3==1.25.11","xgboost>=0.90","torch>=1.8.1",
      "onnxmltools>=1.6.1","Pympler==0.9","docker==5.0.0","wget==3.2","PyJWT==2.2.0","seaborn>=0.11.2"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
    include_package_data=True)     
  


# data wrangling
import pandas as pd
import numpy as np

# ml frameworks
import sklearn
from sklearn.model_selection import GridSearchCV, RandomizedSearchCV
import torch
import xgboost
import tensorflow as tf

# onnx modules
import onnx
import skl2onnx
from skl2onnx import convert_sklearn
import tf2onnx
from torch.onnx import export
from onnx.tools.net_drawer import GetPydotGraph, GetOpNodeProducer
import importlib
import onnxmltools
import onnxruntime as rt


# aims modules
from aimodelshare.aws import run_function_on_lambda, get_aws_client

# os etc
import os
import ast
import tempfile
import json
import re
import pickle
import requests
import sys

from pympler import asizeof
from IPython.core.display import display, HTML, SVG
import absl.logging
import networkx as nx

absl.logging.set_verbosity(absl.logging.ERROR)

def _extract_onnx_metadata(onnx_model, framework):
    '''Extracts model metadata from ONNX file.'''

    # get model graph
    graph = onnx_model.graph

    # initialize metadata dict
    metadata_onnx = {}

    # get input shape
    metadata_onnx["input_shape"] = graph.input[0].type.tensor_type.shape.dim[1].dim_value

    # get output shape
    metadata_onnx["output_shape"] = graph.output[0].type.tensor_type.shape.dim[1].dim_value 
    
    # get layers and activations NEW
    # match layers and nodes and initalizers in sinle object
    # insert here....

    # get layers & activations 
    layer_nodes = ['MatMul', 'Gemm', 'Conv'] #add MaxPool, Transpose, Flatten
    activation_nodes = ['Relu', 'Softmax']

    layers = []
    activations = []

    for op_id, op in enumerate(graph.node):

        if op.op_type in layer_nodes:
            layers.append(op.op_type)
            #if op.op_type == 'MaxPool':
                #activations.append(None)
            
        if op.op_type in activation_nodes:
            activations.append(op.op_type)


    # get shapes and parameters
    layers_shapes = []
    layers_n_params = []

    if framework == 'keras':
        initializer = list(reversed(graph.initializer))
        for layer_id, layer in enumerate(initializer):
            if(len(layer.dims)>= 2):
                layers_shapes.append(layer.dims[1])

                try:
                    n_params = int(np.prod(layer.dims) + initializer[layer_id-1].dims)
                except:
                    n_params = None

                layers_n_params.append(n_params)
                

    elif framework == 'pytorch':
        initializer = graph.initializer
        for layer_id, layer in enumerate(initializer):
            if(len(layer.dims)>= 2):
                layers_shapes.append(layer.dims[0])
                n_params = int(np.prod(layer.dims) + initializer[layer_id-1].dims)
                layers_n_params.append(n_params)


    # get model architecture stats
    model_architecture = {'layers_number': len(layers),
                      'layers_sequence': layers,
                      'layers_summary': {i:layers.count(i) for i in set(layers)},
                      'layers_n_params': layers_n_params,
                      'layers_shapes': layers_shapes,
                      'activations_sequence': activations,
                      'activations_summary': {i:activations.count(i) for i in set(activations)}
                     }

    metadata_onnx["model_architecture"] = model_architecture

    return metadata_onnx



def _misc_to_onnx(model, initial_types, transfer_learning=None,
                    deep_learning=None, task_type=None):

    # generate metadata dict 
    metadata = {}
    
    # placeholders, need to be generated elsewhere
    metadata['model_id'] = None
    metadata['data_id'] = None
    metadata['preprocessor_id'] = None
    
    # infer ml framework from function call
    if isinstance(model, (xgboost.XGBClassifier, xgboost.XGBRegressor)):
        metadata['ml_framework'] = 'xgboost'
        onx = onnxmltools.convert.convert_xgboost(model, initial_types=initial_types)

    # also integrate lightGBM 
    
    # get model type from model object
    model_type = str(model).split('(')[0]
    metadata['model_type'] = model_type
    
    # get transfer learning bool from user input
    metadata['transfer_learning'] = transfer_learning

    # get deep learning bool from user input
    metadata['deep_learning'] = deep_learning
    
    # get task type from user input
    metadata['task_type'] = task_type
    
    # placeholders, need to be inferred from data 
    metadata['target_distribution'] = None
    metadata['input_type'] = None
    metadata['input_shape'] = None
    metadata['input_dtypes'] = None       
    metadata['input_distribution'] = None
    
    # get model config dict from sklearn model object
    metadata['model_config'] = str(model.get_params())

    # get model state from sklearn model object
    metadata['model_state'] = None


    model_architecture = {}

    if hasattr(model, 'coef_'):
        model_architecture['layers_n_params'] = [len(model.coef_.flatten())]
    if hasattr(model, 'solver'):
        model_architecture['optimizer'] = model.solver

    metadata['model_architecture'] = str(model_architecture)

    metadata['memory_size'] = asizeof.asizeof(model)    


    # placeholder, needs evaluation engine
    metadata['eval_metrics'] = None  
    
    # add metadata from onnx object
    # metadata['metadata_onnx'] = str(_extract_onnx_metadata(onx, framework='sklearn'))
    metadata['metadata_onnx'] = None

    meta = onx.metadata_props.add()
    meta.key = 'model_metadata'
    meta.value = str(metadata)

    return onx



def _sklearn_to_onnx(model, initial_types, transfer_learning=None,
                    deep_learning=None, task_type=None):
    '''Extracts metadata from sklearn model object.'''
    
    # check whether this is a fitted sklearn model
    # sklearn.utils.validation.check_is_fitted(model)

    # deal with pipelines and parameter search 
    if isinstance(model, (GridSearchCV, RandomizedSearchCV)):
        model = model.best_estimator_

    if isinstance(model, sklearn.pipeline.Pipeline):
        model = model.steps[-1][1]

    # convert to onnx
    onx = convert_sklearn(model, initial_types=initial_types)
    
    ## Dynamically set model ir_version to ensure sklearn opsets work properly
    from onnx.helper import VERSION_TABLE
    import onnx
    import numpy as np

    indexlocationlist=[]
    for i in VERSION_TABLE:
      indexlocationlist.append(str(i).find(str(onnx.__version__)))


    arr = np.array(indexlocationlist)

    def condition(x): return x > -1

    bool_arr = condition(arr)

    output = np.where(bool_arr)[0]

    ir_version=VERSION_TABLE[output[0]][1]

    #add to model object before saving
    onx.ir_version = ir_version
    
    # generate metadata dict 
    metadata = {}
    
    # placeholders, need to be generated elsewhere
    metadata['model_id'] = None
    metadata['data_id'] = None
    metadata['preprocessor_id'] = None
    
    # infer ml framework from function call
    metadata['ml_framework'] = 'sklearn'
    
    # get model type from model object
    model_type = str(model).split('(')[0]
    metadata['model_type'] = model_type
    
    # get transfer learning bool from user input
    metadata['transfer_learning'] = transfer_learning

    # get deep learning bool from user input
    metadata['deep_learning'] = deep_learning
    
    # get task type from user input
    metadata['task_type'] = task_type
    
    # placeholders, need to be inferred from data 
    metadata['target_distribution'] = None
    metadata['input_type'] = None
    metadata['input_shape'] = None
    metadata['input_dtypes'] = None       
    metadata['input_distribution'] = None
    
    # get model config dict from sklearn model object
    metadata['model_config'] = str(model.get_params())

    # get weights for pretrained models 
    temp_dir = tempfile.gettempdir()
    temp_path = os.path.join(temp_dir, 'temp_file_name')

    with open(temp_path, 'wb') as f:
        pickle.dump(model, f)

    with open(temp_path, "rb") as f:
        pkl_str = f.read()

    metadata['model_weights'] = pkl_str
    
    # get model state from sklearn model object
    metadata['model_state'] = None

    # get model architecture    
    if model_type == 'MLPClassifier' or model_type == 'MLPRegressor':

        if model_type == 'MLPClassifier':
            loss = 'log-loss'
        if model_type == 'MLPRegressor':
            loss = 'squared-loss'

        n_params = []
        layer_dims = [model.n_features_in_] + model.hidden_layer_sizes + [model.n_outputs_]
        for i in range(len(layer_dims)-1):
            n_params.append(layer_dims[i]*layer_dims[i+1] + layer_dims[i+1])

        # insert data into model architecture dict 
        model_architecture = {'layers_number': len(model.hidden_layer_sizes),
                              'layers_sequence': ['Dense']*len(model.hidden_layer_sizes),
                              'layers_summary': {'Dense': len(model.hidden_layer_sizes)},
                              'layers_n_params': n_params, #double check 
                              'layers_shapes': model.hidden_layer_sizes,
                              'activations_sequence': [model.activation]*len(model.hidden_layer_sizes),
                              'activations_summary': {model.activation: len(model.hidden_layer_sizes)},
                              'loss': loss,
                              'optimizer': model.solver
                             }

        metadata['model_architecture'] = str(model_architecture)


    else:
        model_architecture = {}

        if hasattr(model, 'coef_'):
            model_architecture['layers_n_params'] = [len(model.coef_.flatten())]
        if hasattr(model, 'solver'):
            model_architecture['optimizer'] = model.solver

        metadata['model_architecture'] = str(model_architecture)

    metadata['memory_size'] = asizeof.asizeof(model)    

    # placeholder, needs evaluation engine
    metadata['eval_metrics'] = None  
    
    # add metadata from onnx object
    # metadata['metadata_onnx'] = str(_extract_onnx_metadata(onx, framework='sklearn'))
    metadata['metadata_onnx'] = None

    meta = onx.metadata_props.add()
    meta.key = 'model_metadata'
    meta.value = str(metadata)

    return onx



def _keras_to_onnx(model, transfer_learning=None,
                  deep_learning=None, task_type=None, epochs=None):
    '''Extracts metadata from keras model object.'''

    # check whether this is a fitted keras model
    # isinstance...


    # handle keras models in sklearn wrapper
    if isinstance(model, (GridSearchCV, RandomizedSearchCV)):
        model = model.best_estimator_

    if isinstance(model, sklearn.pipeline.Pipeline):
        model = model.steps[-1][1]

    sklearn_wrappers = (tf.keras.wrappers.scikit_learn.KerasClassifier,
                    tf.keras.wrappers.scikit_learn.KerasRegressor)

    if isinstance(model, sklearn_wrappers):
        model = model.model
    
    # convert to onnx
    #onx = convert_keras(model)
    # generate tempfile for onnx object 
    temp_dir = os.path.join(tempfile.gettempdir(), 'test')
    temp_dir = tempfile.gettempdir()



    
    tf.get_logger().setLevel('ERROR') # probably not good practice
    output_path = os.path.join(temp_dir, 'temp.onnx')

    model.save(temp_dir)

    # Convert the model
    converter = tf.lite.TFLiteConverter.from_saved_model(temp_dir) # path to the SavedModel directory
    tflite_model = converter.convert()

    # Save the model.
    with open(os.path.join(temp_dir,'tempmodel.tflite'), 'wb') as f:
      f.write(tflite_model)

    modelstringtest="python -m tf2onnx.convert --tflite "+os.path.join(temp_dir,'tempmodel.tflite')+" --output "+output_path+" --opset 13"
    os.system(modelstringtest)
    output_path = os.path.join(temp_dir, 'temp.onnx')
    modelstringtest="python -m tf2onnx.convert --saved-model "+temp_dir+" --output "+output_path+" --opset 13"
    os.system(modelstringtest)
    onx = onnx.load(output_path)


    # generate metadata dict 
    metadata = {}

    # placeholders, need to be generated elsewhere
    metadata['model_id'] = None
    metadata['data_id'] = None
    metadata['preprocessor_id'] = None

    # infer ml framework from function call
    metadata['ml_framework'] = 'keras'

    # get model type from model object
    metadata['model_type'] =  str(model.__class__.__name__)

    # get transfer learning bool from user input
    metadata['transfer_learning'] = transfer_learning

    # get deep learning bool from user input
    metadata['deep_learning'] = deep_learning

    # get task type from user input
    metadata['task_type'] = task_type

    # placeholders, need to be inferred from data 
    metadata['target_distribution'] = None
    metadata['input_type'] = None
    metadata['input_shape'] = None
    metadata['input_dtypes'] = None       
    metadata['input_distribution'] = None

    # get model config dict from keras model object
    metadata['model_config'] = str(model.get_config())

    # get model weights from keras object 
    def to_list(x):
        return x.tolist()

    metadata['model_weights'] = str(list(map(to_list, model.get_weights())))

    # get model state from pytorch model object
    metadata['model_state'] = None
    
    # get list of current layer types 
    layer_list, activation_list = _get_layer_names()

    # extract model architecture metadata 
    layers = []
    layers_n_params = []
    layers_shapes = []
    activations = []
    
    for i in model.layers: 
        
        # get layer names 
        if i.__class__.__name__ in layer_list:
            layers.append(i.__class__.__name__)
            layers_n_params.append(i.count_params())
            layers_shapes.append(i.output_shape)
        
        # get activation names
        if i.__class__.__name__ in activation_list: 
            activations.append(i.__class__.__name__.lower())
        if hasattr(i, 'activation') and i.activation.__name__ in activation_list:
            activations.append(i.activation.__name__)

    if hasattr(model, 'loss'):
        loss = model.loss.__class__.__name__
    else:
        loss = None
        
    if hasattr(model, 'optimizer'):
        optimizer = model.optimizer.__class__.__name__
    else:
        optimizer = None 

    model_summary_pd = model_summary_keras(model)
    
    # insert data into model architecture dict 
    model_architecture = {'layers_number': len(layers),
                          'layers_sequence': layers,
                          'layers_summary': {i:layers.count(i) for i in set(layers)},
                          'layers_n_params': layers_n_params,
                          'layers_shapes': layers_shapes,
                          'activations_sequence': activations,
                          'activations_summary': {i:activations.count(i) for i in set(activations)},
                          'loss':loss,
                          'optimizer': optimizer
                         }

    metadata['model_architecture'] = str(model_architecture)

    metadata['model_summary'] = model_summary_pd.to_json()

    metadata['memory_size'] = asizeof.asizeof(model)

    metadata['epochs'] = epochs

    # placeholder, needs evaluation engine
    metadata['eval_metrics'] = None

    # add metadata from onnx object
    # metadata['metadata_onnx'] = str(_extract_onnx_metadata(onx, framework='keras'))
    metadata['metadata_onnx'] = None
    # add metadata dict to onnx object
    
    meta = onx.metadata_props.add()
    meta.key = 'model_metadata'
    meta.value = str(metadata)

    return onx


def _pytorch_to_onnx(model, model_input, transfer_learning=None, 
                    deep_learning=None, task_type=None, 
                    epochs=None):
    
    '''Extracts metadata from pytorch model object.'''

    # TODO check whether this is a fitted pytorch model
    # isinstance...

    # generate tempfile for onnx object 
    temp_dir = tempfile.gettempdir()
    temp_path = os.path.join(temp_dir, 'temp_file_name')

    # generate onnx file and save it to temporary path
    export(model, model_input, temp_path)

    # load onnx file from temporary path
    onx = onnx.load(temp_path)

    # generate metadata dict 
    metadata = {}

    # placeholders, need to be generated elsewhere
    metadata['model_id'] = None
    metadata['data_id'] = None
    metadata['preprocessor_id'] = None

    # infer ml framework from function call
    metadata['ml_framework'] = 'pytorch'

    # get model type from model object
    metadata['model_type'] = str(model.__class__).split('.')[-1].split("'")[0] + '()'

    # get transfer learning bool from user input
    metadata['transfer_learning'] = transfer_learning 

    # get deep learning bool from user input
    metadata['deep_learning'] = deep_learning

    # get task type from user input 
    metadata['task_type'] = task_type

    # placeholders, need to be inferred from data 
    metadata['target_distribution'] = None
    metadata['input_type'] = None
    metadata['input_shape'] = None
    metadata['input_dtypes'] = None       
    metadata['input_distribution'] = None

    # get model config dict from pytorch model object
    metadata['model_config'] = str(model.__dict__)

    # get model state from pytorch model object
    metadata['model_state'] = str(model.state_dict())

    # extract model architecture metadata
    activation_names = ['ReLU', 'Softmax']   #placeholder
    layer_names = []

    layers = []
    layers_shapes = []
    n_params = []


    for i, j  in model.__dict__['_modules'].items():

        if j._get_name() not in activation_names:

            layers.append(j._get_name())
            layers_shapes.append(j.out_features)
            weights = np.prod(j._parameters['weight'].shape)
            bias = np.prod(j._parameters['bias'].shape)
            n_params.append(weights+bias)


    activations = []
    for i, j in model.__dict__['_modules'].items():
        if str(j).split('(')[0] in activation_names:
            activations.append(str(j).split('(')[0])


    model_architecture = {'layers_number': len(layers),
                  'layers_sequence': layers,
                  'layers_summary': {i:layers.count(i) for i in set(layers)},
                  'layers_n_params': n_params,
                  'layers_shapes': layers_shapes,
                  'activations_sequence': activations,
                  'activations_summary': {i:activations.count(i) for i in set(activations)}
                 }

    metadata['model_architecture'] = str(model_architecture)

    metadata['memory_size'] = asizeof.asizeof(model)    
    metadata['epochs'] = epochs

    # placeholder, needs evaluation engine
    metadata['eval_metrics'] = None

    # add metadata from onnx object
    # metadata['metadata_onnx'] = str(_extract_onnx_metadata(onx, framework='pytorch'))
    metadata['metadata_onnx'] = None

    # add metadata dict to onnx object
    meta = onx.metadata_props.add()
    meta.key = 'model_metadata'
    meta.value = str(metadata)
    

    return onx


def model_to_onnx(model, framework, model_input=None, initial_types=None,
                  transfer_learning=None, deep_learning=None, task_type=None, 
                  epochs=None):
    
    '''Transforms sklearn, keras, or pytorch model object into ONNX format 
    and extracts model metadata dictionary. The model metadata dictionary 
    is saved in the ONNX file's metadata_props. 
    
    Parameters: 
    model: fitted sklearn, keras, or pytorch model object
    Specifies the model object that will be converted to ONNX. 
    
    framework: {"sklearn", "keras", "pytorch"}
    Specifies the machine learning framework of the model object.
    
    model_input: array_like, default=None
    Required when framework="pytorch".
    initial_types: initial types tuple, default=None
    Required when framework="sklearn".
     
    transfer_learning: bool, default=None
    Indicates whether transfer learning was used. 
    
    deep_learning: bool, default=None
    Indicates whether deep learning was used. 
    
    task_type: {"classification", "regression"}
    Indicates whether the model is a classification model or
    a regression model.
    
    Returns:
    ONNX object with model metadata saved in metadata props
    '''

    # assert that framework exists
    frameworks = ['sklearn', 'keras', 'pytorch', 'xgboost']
    assert framework in frameworks, \
    'Please choose "sklearn", "keras", "pytorch", or "xgboost".'
    
    # assert model input type THIS IS A PLACEHOLDER
    if model_input != None:
        assert isinstance(model_input, (list, pd.core.series.Series, np.ndarray, torch.Tensor)), \
        'Please format model input as XYZ.'
    
    # assert initialtypes 
    if initial_types != None:
        assert isinstance(initial_types[0][1], (skl2onnx.common.data_types.FloatTensorType)), \
        'Please use FloatTensorType as initial types.'
        
    # assert transfer_learning
    if transfer_learning != None:
        assert isinstance(transfer_learning, (bool)), \
        'Please pass boolean to indicate whether transfer learning was used.'
        
    # assert deep_learning
    if deep_learning != None:
        assert isinstance(deep_learning, (bool)), \
        'Please pass boolean to indicate whether deep learning was used.'
        
    # assert task_type
    if task_type != None:
        assert task_type in ['classification', 'regression'], \
        'Please specify task type as "classification" or "regression".'
    

    if framework == 'sklearn':
        onnx = _sklearn_to_onnx(model, initial_types=initial_types, 
                                transfer_learning=transfer_learning, 
                                deep_learning=deep_learning, 
                                task_type=task_type)
    elif framework == 'xgboost':
        onnx = _misc_to_onnx(model, initial_types=initial_types, 
                                transfer_learning=transfer_learning, 
                                deep_learning=deep_learning, 
                                task_type=task_type)
        
    elif framework == 'keras':
        onnx = _keras_to_onnx(model, transfer_learning=transfer_learning, 
                              deep_learning=deep_learning, 
                              task_type=task_type,
                              epochs=epochs)

        
    elif framework == 'pytorch':
        onnx = _pytorch_to_onnx(model, model_input=model_input,
                                transfer_learning=transfer_learning, 
                                deep_learning=deep_learning, 
                                task_type=task_type,
                                epochs=epochs)


    try: 
        rt.InferenceSession(onnx.SerializeToString())   
    except Exception as e: 
        print(e)

    return onnx



def _get_metadata(onnx_model):
    '''Fetches previously extracted model metadata from ONNX object
    and returns model metadata dict.'''
    
    # double check this 
    #assert(isinstance(onnx_model, onnx.onnx_ml_pb2.ModelProto)), \
     #"Please pass a onnx model object."
    
    try: 
        onnx_meta = onnx_model.metadata_props

        onnx_meta_dict = {'model_metadata': ''}

        for i in onnx_meta:
            onnx_meta_dict[i.key] = i.value

        onnx_meta_dict = ast.literal_eval(onnx_meta_dict['model_metadata'])
        
        #if onnx_meta_dict['model_config'] != None and \
        #onnx_meta_dict['ml_framework'] != 'pytorch':
        #    onnx_meta_dict['model_config'] = ast.literal_eval(onnx_meta_dict['model_config'])
        
        if onnx_meta_dict['model_architecture'] != None:
            onnx_meta_dict['model_architecture'] = ast.literal_eval(onnx_meta_dict['model_architecture'])
            
        if onnx_meta_dict['metadata_onnx'] != None:
            onnx_meta_dict['metadata_onnx'] = ast.literal_eval(onnx_meta_dict['metadata_onnx'])
        
        # onnx_meta_dict['model_image'] = onnx_to_image(onnx_model)

    except Exception as e:
    
        print(e)
        
        onnx_meta_dict = ast.literal_eval(onnx_meta_dict)
        
    return onnx_meta_dict



def _get_leaderboard_data(onnx_model, eval_metrics=None):
    
    if eval_metrics is not None:
        metadata = eval_metrics
    else:
        metadata = dict()
        
    metadata_raw = _get_metadata(onnx_model)

    # get list of current layer types 
    layer_list, activation_list = _get_layer_names()

    # get general model info
    metadata['ml_framework'] = metadata_raw['ml_framework']
    metadata['transfer_learning'] = metadata_raw['transfer_learning']
    metadata['deep_learning'] = metadata_raw['deep_learning']
    metadata['model_type'] = metadata_raw['model_type']


    # get neural network metrics
    if metadata_raw['ml_framework'] == 'keras' or metadata_raw['model_type'] in ['MLPClassifier', 'MLPRegressor']:
        metadata['depth'] = metadata_raw['model_architecture']['layers_number']
        metadata['num_params'] = sum(metadata_raw['model_architecture']['layers_n_params'])

        for i in layer_list:
            if i in metadata_raw['model_architecture']['layers_summary']:
                metadata[i.lower()+'_layers'] = metadata_raw['model_architecture']['layers_summary'][i]
            else:
                metadata[i.lower()+'_layers'] = 0

        for i in activation_list:
            if i in metadata_raw['model_architecture']['activations_summary']:
                if i.lower()+'_act' in metadata:
                    metadata[i.lower()+'_act'] += metadata_raw['model_architecture']['activations_summary'][i]
                else:    
                    metadata[i.lower()+'_act'] = metadata_raw['model_architecture']['activations_summary'][i]
            else:
                if i.lower()+'_act' not in metadata:
                    metadata[i.lower()+'_act'] = 0
                
        metadata['loss'] = metadata_raw['model_architecture']['loss']
        metadata['optimizer'] = metadata_raw['model_architecture']["optimizer"]
        metadata['model_config'] = metadata_raw['model_config']
        metadata['epochs'] = metadata_raw['epochs']
        metadata['memory_size'] = metadata_raw['memory_size']



    # get sklearn model metrics
    elif metadata_raw['ml_framework'] == 'sklearn' or metadata_raw['ml_framework'] == 'xgboost':
        metadata['depth'] = 0

        try:
            metadata['num_params'] = sum(metadata_raw['model_architecture']['layers_n_params'])
        except:
            metadata['num_params'] = 0

        for i in layer_list:
            metadata[i.lower()+'_layers'] = 0

        for i in activation_list:
            metadata[i.lower()+'_act'] = 0

        metadata['loss'] = None

        try:
            metadata['optimizer'] = metadata_raw['model_architecture']['optimizer']
        except:
            metadata['optimizer'] = None

        try:
            metadata['model_config'] = metadata_raw['model_config']
        except:
            metadata['model_config'] = None
    
    return metadata
    


def _model_summary(meta_dict, from_onnx=False):
    '''Creates model summary table from model metadata dict.'''
    
    assert(isinstance(meta_dict, dict)), \
    "Please pass valid metadata dict."
    
    assert('model_architecture' in meta_dict.keys()), \
    "Please make sure model architecture data is included."

    if from_onnx == True:
        model_summary = pd.read_json(meta_dict['metadata_onnx']["model_summary"])
    else:
        model_summary = pd.read_json(meta_dict["model_summary"])
       
    return model_summary



def onnx_to_image(model):
    '''Creates model graph image in pydot format.'''
    
    OP_STYLE = {
    'shape': 'box',
    'color': '#0F9D58',
    'style': 'filled',
    'fontcolor': '#FFFFFF'
    }
    
    pydot_graph = GetPydotGraph(
        model.graph,
        name=model.graph.name,
        rankdir='TB',
        node_producer=GetOpNodeProducer(
            embed_docstring=False,
            **OP_STYLE
        )
    )
    
    return pydot_graph

def inspect_model_aws(apiurl, version=None):
    if all(["AWS_ACCESS_KEY_ID" in os.environ, 
            "AWS_SECRET_ACCESS_KEY" in os.environ,
            "AWS_REGION" in os.environ, 
           "username" in os.environ, 
           "password" in os.environ]):
        pass
    else:
        return print("'Inspect Model' unsuccessful. Please provide credentials with set_credentials().")

    aws_client = get_aws_client() 

    onnx_model = _get_onnx_from_bucket(apiurl, aws_client, version=version)

    meta_dict = _get_metadata(onnx_model)

    if meta_dict['ml_framework'] == 'keras':
        inspect_pd = _model_summary(meta_dict)
        
    elif meta_dict['ml_framework'] in ['sklearn', 'xgboost']:
        model_config = meta_dict["model_config"]
        model_config = ast.literal_eval(model_config)
        inspect_pd = pd.DataFrame({'param_name': model_config.keys(),
                                   'param_value': model_config.values()})
    
    return inspect_pd

def inspect_model_lambda(apiurl, version=None):
    if all(["username" in os.environ, 
           "password" in os.environ]):
        pass
    else:
        return print("'Inspect Model' unsuccessful. Please provide credentials with set_credentials().")

    post_dict = {"y_pred": [],
               "return_eval": "False",
               "return_y": "False",
               "inspect_model": "True",
               "version": version}
    
    headers = { 'Content-Type':'application/json', 'authorizationToken': os.environ.get("AWS_TOKEN"),} 

    apiurl_eval=apiurl[:-1]+"eval"

    inspect_json = requests.post(apiurl_eval,headers=headers,data=json.dumps(post_dict)) 

    inspect_pd = pd.DataFrame(json.loads(inspect_json.text))

    return inspect_pd


def inspect_model_dict(apiurl, version=None):

    if all(["AWS_ACCESS_KEY_ID" in os.environ, 
            "AWS_SECRET_ACCESS_KEY" in os.environ,
            "AWS_REGION" in os.environ, 
           "username" in os.environ, 
           "password" in os.environ]):
        pass
    else:
        return print("'Inspect Model' unsuccessful. Please provide credentials with set_credentials().")

    aws_client=get_aws_client(aws_key=os.environ.get('AWS_ACCESS_KEY_ID'), 
                              aws_secret=os.environ.get('AWS_SECRET_ACCESS_KEY'), 
                              aws_region=os.environ.get('AWS_REGION'))

    # Get bucket and model_id for user
    response, error = run_function_on_lambda(
        apiurl, **{"delete": "FALSE", "versionupdateget": "TRUE"}
    )
    if error is not None:
        raise error

    _, bucket, model_id = json.loads(response.content.decode("utf-8"))

    key = model_id+'/inspect_pd.json'
    
    try:
      resp = aws_client['client'].get_object(Bucket=bucket, Key=key)
      data = resp.get('Body').read()
      model_dict = json.loads(data)
    except Exception as e:
        print(e)

    ml_framework = model_dict.get(str(version))['ml_framework']
    model_type = model_dict.get(str(version))['model_type']
    inspect_pd = pd.DataFrame(model_dict.get(str(version))['model_dict'])
    
    return inspect_pd



def inspect_model(apiurl, version=None):
    if all(["username" in os.environ, 
           "password" in os.environ]):
        pass
    else:
        return print("'Inspect Model' unsuccessful. Please provide credentials with set_credentials().")

    try:
        inspect_pd = inspect_model_lambda(apiurl, version)
    except:

        try: 
            inspect_pd = inspect_model_dict(apiurl, version)
        except: 
            inspect_pd = inspect_model_aws(apiurl, version)
    
    return inspect_pd


def compare_models_dict(apiurl, version_list=None, 
    by_model_type=None, best_model=None, verbose=1):
    
    if not isinstance(version_list, list):
        raise Exception("Argument 'version' must be a list.")
    
    if all(["AWS_ACCESS_KEY_ID" in os.environ, 
            "AWS_SECRET_ACCESS_KEY" in os.environ,
            "AWS_REGION" in os.environ, 
           "username" in os.environ, 
           "password" in os.environ]):
        pass
    else:
        return print("'Compare Models' unsuccessful. Please provide credentials with set_credentials().")

    aws_client=get_aws_client(aws_key=os.environ.get('AWS_ACCESS_KEY_ID'), 
                              aws_secret=os.environ.get('AWS_SECRET_ACCESS_KEY'), 
                              aws_region=os.environ.get('AWS_REGION'))

    # Get bucket and model_id for user
    response, error = run_function_on_lambda(
        apiurl, **{"delete": "FALSE", "versionupdateget": "TRUE"}
    )
    if error is not None:
        raise error

    _, bucket, model_id = json.loads(response.content.decode("utf-8"))

    key = model_id+'/inspect_pd.json'
    
    try:
      resp = aws_client['client'].get_object(Bucket=bucket, Key=key)
      data = resp.get('Body').read()
      model_dict = json.loads(data)
    except Exception as e:
        print(e)

    ml_framework_list = [model_dict[str(i)]['ml_framework'] for i in version_list]
    model_type_list = [model_dict[str(i)]['model_type'] for i in version_list]
    model_dict_list = [model_dict[str(i)]['model_dict'] for i in version_list]

    if not all(x==ml_framework_list[0] for x in ml_framework_list):
        raise Exception("Incongruent frameworks. Please compare models from the same ML frameworks.")
        

    if ml_framework_list[0] == 'sklearn':

        comp_dict_out = {}
        
        for i in version_list: 
            
            temp_pd = pd.DataFrame(model_dict[str(i)]['model_dict'])
            temp_pd.columns = ['param_name', 'default_value', "model_version_"+str(i)]

            if model_dict[str(i)]['model_type'] in comp_dict_out.keys():

                comp_pd = comp_dict_out[model_dict[str(i)]['model_type']]
                comp_pd = comp_pd.merge(temp_pd.drop('default_value', axis=1), on='param_name')

                comp_dict_out[model_dict[str(i)]['model_type']] = comp_pd

            else:
                comp_dict_out[model_dict[str(i)]['model_type']] = temp_pd

        return comp_dict_out

            
    if ml_framework_list[0] == 'keras':

        comp_pd = pd.DataFrame()

        for i in version_list: 

            temp_pd = pd.DataFrame(model_dict[str(i)]['model_dict'])

            temp_pd.iloc[:,2] = temp_pd.iloc[:,2].astype(str)

            if verbose == 0: 
                temp_pd = temp_pd[['Layer']]
            elif verbose == 1: 
                temp_pd = temp_pd[['Layer', 'Shape', 'Params']]
            elif verbose == 2: 
                temp_pd = temp_pd[['Name', 'Layer', 'Shape', 'Params', 'Connect']]
            elif verbose == 3: 
                temp_pd = temp_pd[['Name', 'Layer', 'Shape', 'Params', 'Connect', 'Activation']]

            temp_pd = temp_pd.add_prefix('Model_'+str(i)+'_')    
            comp_pd = pd.concat([comp_pd, temp_pd], axis=1)

        comp_dict_out = {"Sequential": comp_pd}
        
        return comp_dict_out


def color_pal_assign(val):
  import pandas as pd
  
  # create Pandas Series with default index values
  # default index ranges is from 0 to len(list) - 1
  layer_name_df =  pd.DataFrame(['AbstractRNNCell',
  'Activation',
  'ActivityRegularization',
  'Add',
  'AdditiveAttention',
  'AlphaDropout',
  'Attention',
  'Average',
  'AveragePooling1D',
  'AveragePooling2D',
  'AveragePooling3D',
  'AvgPool1D',
  'AvgPool2D',
  'AvgPool3D',
  'BatchNormalization',
  'Bidirectional',
  'CategoryEncoding',
  'CenterCrop',
  'Concatenate',
  'Conv1D',
  'Conv1DTranspose',
  'Conv2D',
  'Conv2DTranspose',
  'Conv3D',
  'Conv3DTranspose',
  'ConvLSTM1D',
  'ConvLSTM2D',
  'ConvLSTM3D',
  'Convolution1D',
  'Convolution1DTranspose',
  'Convolution2D',
  'Convolution2DTranspose',
  'Convolution3D',
  'Convolution3DTranspose',
  'Cropping1D',
  'Cropping2D',
  'Cropping3D',
  'Dense',
  'DenseFeatures',
  'DepthwiseConv2D',
  'Discretization',
  'Dot',
  'Dropout',
  'Embedding',
  'Flatten',
  'GRU',
  'GRUCell',
  'GaussianDropout',
  'GaussianNoise',
  'GlobalAveragePooling1D',
  'GlobalAveragePooling2D',
  'GlobalAveragePooling3D',
  'GlobalAvgPool1D',
  'GlobalAvgPool2D',
  'GlobalAvgPool3D',
  'GlobalMaxPool1D',
  'GlobalMaxPool2D',
  'GlobalMaxPool3D',
  'GlobalMaxPooling1D',
  'GlobalMaxPooling2D',
  'GlobalMaxPooling3D',
  'Hashing',
  'Input',
  'InputLayer',
  'InputSpec',
  'IntegerLookup',
  'LSTM',
  'LSTMCell',
  'Lambda',
  'Layer',
  'LayerNormalization',
  'LocallyConnected1D',
  'LocallyConnected2D',
  'Masking',
  'MaxPool1D',
  'MaxPool2D',
  'MaxPool3D',
  'MaxPooling1D',
  'MaxPooling2D',
  'MaxPooling3D',
  'Maximum',
  'Minimum',
  'MultiHeadAttention',
  'Multiply',
  'Normalization',
  'Permute',
  'RNN',
  'RandomContrast',
  'RandomCrop',
  'RandomFlip',
  'RandomHeight',
  'RandomRotation',
  'RandomTranslation',
  'RandomWidth',
  'RandomZoom',
  'RepeatVector',
  'Rescaling',
  'Reshape',
  'Resizing',
  'SeparableConv1D',
  'SeparableConv2D',
  'SeparableConvolution1D',
  'SeparableConvolution2D',
  'SimpleRNN',
  'SimpleRNNCell',
  'SpatialDropout1D',
  'SpatialDropout2D',
  'SpatialDropout3D',
  'StackedRNNCells',
  'StringLookup',
  'Subtract',
  'TextVectorization',
  'TimeDistributed',
  'UpSampling1D',
  'UpSampling2D',
  'UpSampling3D',
  'Wrapper',
  'ZeroPadding1D',
  'ZeroPadding2D',
  'ZeroPadding3D'])

  layernamelist=list(layer_name_df[0])
  import seaborn as sns
  paldata=sns.color_palette("Pastel2", len(layernamelist)).as_hex()

  if val in layernamelist: 
      valindex=layernamelist.index(val)
      if any([val=="Concatenate",val=="Conv2D"]):
        valindex=valindex+4
      else:
        pass
      palvalue=paldata[valindex]
  else:
     pass
  color = palvalue if val in layernamelist else 'white'
  return 'background: %s' % color

def stylize_model_comparison(comp_dict_out):


    if 'Sequential' in comp_dict_out.keys():

        df_styled = comp_dict_out['Sequential'].style.applymap(color_pal_assign)

        df_styled = df_styled.set_properties(**{'color': 'black'})

        df_styled = df_styled.set_caption('Model type: ' + 'Neural Network').set_table_styles([{'selector': 'caption',
            'props': [('color', 'white'), ('font-size', '18px')]}])

        df_styled = df_styled.set_properties(**{'color': 'black'})

        df_styled = df_styled.set_caption('Model type: ' + 'Neural Network').set_table_styles([{'selector': 'caption',
            'props': [('color', 'black'), ('font-size', '18px')]}])

        display(HTML(df_styled.render()))


    else:

        for i in comp_dict_out.keys():

            comp_pd = comp_dict_out[i]

            df_styled = comp_pd.style.apply(lambda x: ["background: tomato" if v != x.iloc[0] else "" for v in x], 
                axis = 1, subset=comp_pd.columns[1:])

            df_styled = df_styled.set_caption('Model type: ' + i).set_table_styles([{'selector': 'caption',
                'props': [('color', 'black'), ('font-size', '18px')]}])

            display(HTML(df_styled.render()))
            print('\n\n')


def compare_models_aws(apiurl, version_list=None, 
    by_model_type=None, best_model=None, verbose=3):
    
    if not isinstance(version_list, list):
        raise Exception("Argument 'version' must be a list.")
    
    if all(["AWS_ACCESS_KEY_ID" in os.environ, 
            "AWS_SECRET_ACCESS_KEY" in os.environ,
            "AWS_REGION" in os.environ, 
           "username" in os.environ, 
           "password" in os.environ]):
        pass
    else:
        return print("'Compare Models' unsuccessful. Please provide credentials with set_credentials().")

    aws_client = get_aws_client()
    
    models_dict = {}
    
    for i in version_list: 
        
        onnx_model = _get_onnx_from_bucket(apiurl, aws_client, version=i)
        meta_dict = _get_metadata(onnx_model)
        
        models_dict[i] = meta_dict
        
    ml_framework_list = [models_dict[i]['ml_framework'] for i in models_dict]
    model_type_list = [models_dict[i]['model_type'] for i in models_dict]
    
    if not all(x==ml_framework_list[0] for x in ml_framework_list):
        raise Exception("Incongruent frameworks. Please compare models from the same ML frameworks.")
        
    if not all(x==model_type_list[0] for x in model_type_list):
        raise Exception("Incongruent model types. Please compare models of the same model type.")
    
    if ml_framework_list[0] == 'sklearn':
        
        model_type = model_type_list[0]
        model_class = model_from_string(model_type)
        default = model_class()
        default_config = default.get_params()
        
        comp_pd = pd.DataFrame({'param_name': default_config.keys(),
                           'param_value': default_config.values()})
        
        for i in version_list: 
            
            temp_pd = inspect_model(apiurl, version=i)
            comp_pd = comp_pd.merge(temp_pd, on='param_name')
        
        comp_pd.columns = ['param_name', 'model_default'] + ["Model_"+str(i) for i in version_list]
        
        df_styled = comp_pd.style.apply(lambda x: ["background: tomato" if v != x.iloc[0] else "" for v in x], 
                                        axis = 1, subset=comp_pd.columns[1:])

        
    if ml_framework_list[0] == 'keras':

        comp_pd = pd.DataFrame()

        for i in version_list: 

            temp_pd = inspect_model(apiurl, version=i)

            temp_pd = temp_pd.iloc[:,0:verbose]

            temp_pd = temp_pd.add_prefix('Model_'+str(i)+'_')    
            comp_pd = pd.concat([comp_pd, temp_pd], axis=1, ignore_index=True)

        layer_names = _get_layer_names()

        dense_layers = [i for i in layer_names[0] if 'Dense' in i]
        df_styled = df_styled.style.apply(lambda x: ["background: #DFFF00" if v in dense_layers else "" for v in x], 
                                axis = 1)
        drop_layers = [i for i in layer_names[0] if 'Dropout' in i]
        df_styled = df_styled.apply(lambda x: ["background: #FFBF00" if v in drop_layers else "" for v in x], 
                                axis = 1)
        conv_layers = [i for i in layer_names[0] if 'Conv' in i]
        df_styled = df_styled.apply(lambda x: ["background: #FF7F50" if v in conv_layers else "" for v in x], 
                                axis = 1)
        seq_layers = [i for i in layer_names[0] if 'RNN' in i or 'LSTM' in i or 'GRU' in i] + ['Bidirectional']
        df_styled = df_styled.apply(lambda x: ["background: #DE3163" if v in seq_layers else "" for v in x], 
                                axis = 1)
        pool_layers = [i for i in layer_names[0] if 'Pool' in i]
        df_styled = df_styled.apply(lambda x: ["background: #9FE2BF" if v in pool_layers else "" for v in x], 
                                axis = 1)
        rest_layers = [i for i in layer_names[0] if i not in dense_layers+drop_layers+conv_layers+seq_layers+pool_layers]
        df_styled = df_styled.apply(lambda x: ["background: lightgrey" if v in rest_layers else "" for v in x], 
                            axis = 1)

    df_styled = df_styled.style.set_properties(**{'color': 'lawngreen'})

    return df_styled


def compare_models_lambda(apiurl, version_list="None", 
    by_model_type=None, best_model=None, verbose=1):
    if all(["username" in os.environ, 
           "password" in os.environ]):
        pass
    else:
        return print("'Inspect Model' unsuccessful. Please provide credentials with set_credentials().")

    post_dict = {"y_pred": [],
               "return_eval": "False",
               "return_y": "False",
               "inspect_model": "False",
               "version": "None", 
               "compare_models": "True",
               "version_list": version_list,
               "verbose": verbose}
    
    headers = { 'Content-Type':'application/json', 'authorizationToken': os.environ.get("AWS_TOKEN"),} 

    apiurl_eval=apiurl[:-1]+"eval"

    compare_json = requests.post(apiurl_eval,headers=headers,data=json.dumps(post_dict)) 

    compare_dict = json.loads(compare_json.text)

    comp_dict_out = {i: pd.DataFrame(json.loads(compare_dict[i])) for i in compare_dict}

    return comp_dict_out


def compare_models(apiurl, version_list="None", 
    by_model_type=None, best_model=None, verbose=3):

    if all(["username" in os.environ, 
           "password" in os.environ]):
        pass
    else:
        return print("'Inspect Model' unsuccessful. Please provide credentials with set_credentials().")

    if len(version_list) != len(set(version_list)):
        return print("Model comparison failed. Version list contains duplicates.")
    
    try: 
        compare_pd = compare_models_lambda(apiurl, version_list, 
            by_model_type, best_model, verbose)

    except: 

        try: 
            compare_pd = compare_models_dict(apiurl, version_list, 
            by_model_type, best_model, verbose)

        except:

            compare_pd = compare_models_aws(apiurl, version_list, 
                by_model_type, best_model, verbose)
    
    return compare_pd



def _get_onnx_from_bucket(apiurl, aws_client, version=None):

    # generate name of onnx model in bucket
    onnx_model_name = "/onnx_model_v{version}.onnx".format(version = version)

    # Get bucket and model_id for user
    response, error = run_function_on_lambda(
        apiurl, **{"delete": "FALSE", "versionupdateget": "TRUE"}
    )
    if error is not None:
        raise error

    _, bucket, model_id = json.loads(response.content.decode("utf-8"))

    try:
        onnx_string = aws_client["client"].get_object(
            Bucket=bucket, Key=model_id + onnx_model_name
        )

        onnx_string = onnx_string['Body'].read()

    except Exception as err:
        raise err

    # generate tempfile for onnx object 
    temp_dir = tempfile.gettempdir()
    temp_path = os.path.join(temp_dir, 'temp_file_name')

    # save onnx to temporary path
    with open(temp_path, "wb") as f:
        f.write(onnx_string)

    # load onnx file from temporary path
    onx = onnx.load(temp_path)

    return onx


def instantiate_model(apiurl, version=None, trained=False):
    if all(["AWS_ACCESS_KEY_ID" in os.environ, 
            "AWS_SECRET_ACCESS_KEY" in os.environ,
            "AWS_REGION" in os.environ, 
            "username" in os.environ, 
            "password" in os.environ]):
        pass
    else:
        return print("'Instantiate Model' unsuccessful. Please provide credentials with set_credentials().")

    aws_client = get_aws_client()   
    onnx_model = _get_onnx_from_bucket(apiurl, aws_client, version=version)
    meta_dict = _get_metadata(onnx_model)

    # get model config 
    model_config = ast.literal_eval(meta_dict['model_config'])
    ml_framework = meta_dict['ml_framework']
    
    if ml_framework == 'sklearn':

        if trained == False:
            model_type = meta_dict['model_type']
            model_class = model_from_string(model_type)
            model = model_class(**model_config)

        if trained == True:
            
            model_pkl = meta_dict['model_weights']

            temp_dir = tempfile.gettempdir()
            temp_path = os.path.join(temp_dir, 'temp_file_name')

            with open(temp_path, "wb") as f:
                f.write(model_pkl)

            with open(temp_path, 'rb') as f:
                model = pickle.load(f)


    if ml_framework == 'keras':

        if trained == False:
            model = tf.keras.Sequential().from_config(model_config)

        if trained == True: 
            model = tf.keras.Sequential().from_config(model_config)
            model_weights = json.loads(meta_dict['model_weights'])

            def to_array(x):
                return np.array(x, dtype="float32")

            model_weights = list(map(to_array, model_weights))

            model.set_weights(model_weights)

    return model


def _get_layer_names():

    activation_list = [i for i in dir(tf.keras.activations)]
    activation_list = [i for i in activation_list if callable(getattr(tf.keras.activations, i))]
    activation_list = [i for i in activation_list if  not i.startswith("_")]
    activation_list.remove('deserialize')
    activation_list.remove('get')
    activation_list.remove('linear')
    activation_list = activation_list+['ReLU', 'Softmax', 'LeakyReLU', 'PReLU', 'ELU', 'ThresholdedReLU']


    layer_list = [i for i in dir(tf.keras.layers)]
    layer_list = [i for i in dir(tf.keras.layers) if callable(getattr(tf.keras.layers, i))]
    layer_list = [i for i in layer_list if not i.startswith("_")]
    layer_list = [i for i in layer_list if re.match('^[A-Z]', i)]
    layer_list = [i for i in layer_list if i.lower() not in [i.lower() for i in activation_list]]

    return layer_list, activation_list


def _get_sklearn_modules():
    
    import sklearn

    sklearn_modules = ['ensemble', 'gaussian_process', 'isotonic',
                       'linear_model', 'mixture', 'multiclass', 'naive_bayes', 
                       'neighbors', 'neural_network', 'svm', 'tree']

    models_modules_dict = {}

    for i in sklearn_modules:
        models_list = [j for j in dir(eval('sklearn.'+i)) if callable(getattr(eval('sklearn.'+i), j))]
        models_list = [j for j in models_list if re.match('^[A-Z]', j)]

        for k in models_list: 
            models_modules_dict[k] = 'sklearn.'+i
    
    return models_modules_dict



def model_from_string(model_type):
    models_modules_dict = _get_sklearn_modules()
    module = models_modules_dict[model_type]
    model_class = getattr(importlib.import_module(module), model_type)
    return model_class


def print_y_stats(y_stats): 

  print("y_test example: ", y_stats['ytest_example'])
  print("y_test class labels", y_stats['class_labels'])
  print("y_test class balance", y_stats['class_balance'])
  print("y_test label dtypes", y_stats['label_dtypes'])


def inspect_y_test(apiurl):

  # Confirm that creds are loaded, print warning if not
  if all(["username" in os.environ, 
          "password" in os.environ]):
      pass
  else:
      return print("'Submit Model' unsuccessful. Please provide credentials with set_credentials().")

  post_dict = {"y_pred": [],
               "return_eval": "False",
               "return_y": "True"}
    
  headers = { 'Content-Type':'application/json', 'authorizationToken': os.environ.get("AWS_TOKEN"),} 

  apiurl_eval=apiurl[:-1]+"eval"

  y_stats = requests.post(apiurl_eval,headers=headers,data=json.dumps(post_dict)) 

  y_stats_dict = json.loads(y_stats.text)

  # print_y_stats(y_stats_dict)

  return y_stats_dict


def model_summary_keras(model):

    # extract model architecture metadata 
    layer_names = []
    layer_types = []
    layer_n_params = []
    layer_shapes = []
    layer_connect = []
    activations = []


    for i in model.layers: 

        try:
            layer_names.append(i.name)
        except:
            layer_names.append(None)

        try:
            layer_types.append(i.__class__.__name__)
        except:
            layer_types.append(None)

        try:
            layer_shapes.append(i.output_shape)
        except:
            layer_shapes.append(None)

        try:
            layer_n_params.append(i.count_params())
        except:
            layer_n_params.append(None)

        try:
            if isinstance(i.inbound_nodes[0].inbound_layers, list):
              layer_connect.append([x.name for x in i.inbound_nodes[0].inbound_layers])
            else: 
              layer_connect.append(i.inbound_nodes[0].inbound_layers.name)
        except:
            layer_connect.append(None)

        try: 
            activations.append(i.activation.__name__)
        except:
            activations.append(None)


    model_summary = pd.DataFrame({"Name": layer_names,
    "Layer": layer_types,
    "Shape": layer_shapes,
    "Params": layer_n_params,
    "Connect": layer_connect,
    "Activation": activations})

    return model_summary


def model_graph_keras(model):

    # extract model architecture metadata 
    layer_names = []
    layer_types = []
    layer_n_params = []
    layer_shapes = []
    layer_connect = []
    activations = []

    graph_nodes = []
    graph_edges = []


    for i in model.layers: 

        try:
            layer_name = i.name
        except:
            layer_name = None
        finally:
            layer_names.append(layer_name)


        try:
            layer_type = i.__class__.__name__
        except:
            layer_type = None
        finally:
            layer_types.append(layer_type)

        try:
            layer_shape = i.output_shape
        except:
            layer_shape = None
        finally:
            layer_shapes.append(layer_shape)


        try:
            layer_params = i.count_params()
        except:
            layer_params = None
        finally:
            layer_n_params.append(layer_params)


        try:
            if isinstance(i.inbound_nodes[0].inbound_layers, list):
              layer_input = [x.name for x in i.inbound_nodes[0].inbound_layers]
            else: 
              layer_input = i.inbound_nodes[0].inbound_layers.name
        except:
            layer_connect = None
        finally:
            layer_connect.append(layer_input)


        try: 
            activation = i.activation.__name__
        except:
            activation = None
        finally:
            activations.append(activation)

        layer_color = color_pal_assign(layer_type)
        layer_color =  layer_color.split(' ')[-1]


        graph_nodes.append((layer_name, {"label": layer_type + '\n' + str(layer_shape),
                                         "URL": "https://keras.io/search.html?query="+layer_type.lower(),
                                         "color": layer_color,
                                         "style": "bold",
                                    "Name": layer_name,
                                    "Layer": layer_type,
                                    "Shape": layer_shape,
                                    "Params": layer_params,
                                    "Activation": activation}))
        
        if isinstance(layer_input, list):
            for i in layer_input:
                graph_edges.append((i, layer_name))
        else:
            graph_edges.append((layer_input, layer_name))

    G = nx.DiGraph()
    G.add_nodes_from(graph_nodes)
    G.add_edges_from(graph_edges)

    G_pydot = nx.drawing.nx_pydot.to_pydot(G)

    return G_pydot


def plot_keras(model):

    G =  model_graph_keras(model)

    display(SVG(G.create_svg()))
